'''Library supported features discovering through code inspection.'''

import importlib
from collections import OrderedDict

from http_request_codegen.hrc_factory import (
    get_func_by_lang_impl_method,
    get_generators_modules_by_lang_impl
)
from http_request_codegen.hrc_http import HTTP_METHODS
from http_request_codegen.hrc_meta import function_has_kwarg


FEATURES_KWARGS = OrderedDict({
    'Headers': 'headers',
    'Parameters': 'parameters',
    'Parameters localization': 'locale',
    'Parameters seed': 'seed',
    'Custom indentation': 'indent',
    'Quotation character': 'quote_char',
    'One line rendering': 'oneline',
    'Custom initialization': 'setup',
    'Custom teardown': 'teardown',
    'Line wrapping': 'wrap',
})


def _func_has_kwarg_support_factory(kwarg_name):
    def _func_has_kwarg(language, impl, method):
        func = get_func_by_lang_impl_method(
            language=language, impl=impl, method=method)
        return function_has_kwarg(func, kwarg_name)
    return _func_has_kwarg


def _get_supports_parameters_funcs():
    return OrderedDict({
        feature: _func_has_kwarg_support_factory(kwarg) for feature, kwarg in
        FEATURES_KWARGS.items()
    })


def supported_features():
    '''Returns all supported features for each implementation of the library
    by methods, following the next form:

    ```python
    {
        'language/platform': {
            'implementation_A': {
                'GET': {
                    'Feature-A': True,
                    'Feature-B': False,
                    'Feature-C': False,
                },
                'POST': {
                    'Feature-A': False,
                    'Feature-B': True,
                    'Feature-C': False,
                },
            }
        }
    }
    ```

    A feature is supported is a function that reproduces a method in their
    implementation has a certain keyword argument in his definition. The
    mapping of features and keyword arguments are defined in the global
    variable ``FEATURES_KWARGS`` of this module.

    This function is used to build the "Support" section of the documentation.

    Returns:
        dict: Mapping with all features supported by method for each
            implementation.
    '''
    _supports_params_funcs = _get_supports_parameters_funcs()

    response = {}

    for lang, impls in get_generators_modules_by_lang_impl().items():
        response[lang] = {}
        for impl, module_path in impls.items():
            response[lang][impl] = {}
            mod = importlib.import_module(module_path)

            for method in HTTP_METHODS:
                _method_supported = False

                try:
                    getattr(mod, method.lower())
                except AttributeError:
                    continue
                response[lang][impl][method] = OrderedDict({})
                for feature, support_func in _supports_params_funcs.items():
                    supported = support_func(lang, impl, method)
                    response[lang][impl][method][feature] = supported
                    if supported:
                        _method_supported = True
                response[lang][impl][method]['_supported'] = _method_supported

    return response


def supported_methods():
    '''Returns a dictionary with all languages with all implmentations and
    methods supported for each implementation, in next form:

    ```python
    {
        'language/platform': {
            'implementation_A': ['GET', 'POST', 'PATCH'],
            'implementation_B': ['GET'],
        }
    }
    ```

    A method will be only included if at least one of the required features is
    supported.

    Returns:
        dict: Mapping with all supported methods for each implementation.
    '''
    response = {}
    for lang, impls in supported_features().items():
        response[lang] = {}
        for impl, methods in impls.items():
            _impl_methods = []
            for method, features in methods.items():
                if not features['_supported']:
                    continue
                _impl_methods.append(method)
            if _impl_methods:
                response[lang][impl] = _impl_methods
    return response



from http_request_codegen.hrc_api import (
    generate_http_request_code,
    generate_http_request_md_fenced_code_block
)
from http_request_codegen.hrc_support import (
    supported_features,
    supported_methods
)
from http_request_codegen.hrc_valuer import (
    lazy_name_by_parameter,
    lazy_value_by_parameter
)


__version__ = '0.0.4'
__version_info__ = tuple([int(i) for i in __version__.split('.')])
__title__ = 'http-request-codegen'
__description__ = 'Multilanguage HTTP requests code generator.'
__all__ = (
    'generate_http_request_code',
    'generate_http_request_md_fenced_code_block',
    'lazy_name_by_parameter',
    'lazy_value_by_parameter',
    'supported_features',
    'supported_methods',
)



"""http-request-codegen string utilities."""

import importlib
import random
from collections.abc import Iterable
from types import GeneratorType

from http_request_codegen.hrc_meta import CallableTypes


def lazy_string(string, seed=None, string_func_path=False):
    """Builds a string given an iterable, a callable or the string itself (in
    this case does not transforms it). Useful to randomize a string following
    multiple strategies. Using ``string_func_path``, takes the string as
    a module-callable/iterable path.

    This function is recursive, so if you passes callables as result of your
    callable, make sure that one of them returns other non callable object
    or a ``RecursionError`` will be raised.

    Args:
        string (str/iterable/callable): String or possibilities of strings that
            can be used to build the output. Any other type like numbers or
            booleans will return the ``__repr__`` method of the object.
        seed (int): Seed used in randomization calls.
        string_func_path (bool): If ``True`` and ``string`` is of type ``str``,
            will try to import an object of a module assuming that the string
            has the form ``'path.to.module::object'``. Will raise a
            ``ModuleNotFoundError`` if the module can't be imported or a
            ``ImportError`` if the object to import does not exists in the
            module.

    Examples:
        >>> # String input
        >>> lazy_string('foo')
        'foo'

        >>> # List input
        >>> result = lazy_string(['foo', 'bar'])
        >>> result in ('foo', 'bar')
        True

        >>> # Set input
        >>> result = lazy_string({'foo', 'bar'})
        >>> result in ('foo', 'bar')
        True

        >>> # Tuple input
        >>> result = lazy_string(('foo', 'bar'))
        >>> result in ('foo', 'bar')
        True

        >>> # Generator input
        >>> result = lazy_string((s for s in ('foo', 'bar')))
        >>> result in ('foo', 'bar')
        True

        >>> # Function input
        >>> def func():
        ...     return ['foo', 'bar']
        >>> result = lazy_string(func)
        >>> result in ('foo', 'bar')
        True

        >>> # Function that returns itself
        >>> def func_returning_itself():
        ...     return func_returning_itself
        >>> result = lazy_string(func_returning_itself)
        Traceback (most recent call last):
          ...
        RecursionError: maximum recursion depth exceeded in comparison

        >>> # Lambda input
        >>> result = lazy_string(lambda: ('foo', 'bar'))
        >>> result in ('foo', 'bar')
        True

        >>> # Method input
        >>> class Foo(object):
        ...     def bar(self):
        ...         return ('foo', 'bar')
        >>> result = lazy_string(Foo().bar)
        >>> result in ('foo', 'bar')
        True

        >>> # Boolean input
        >>> lazy_string(False)
        'False'

        >>> # Integer input
        >>> lazy_string(1)
        '1'

        >>> # Float input
        >>> lazy_string(5.73)
        '5.73'

        >>> # Type input
        >>> lazy_string(str)
        'str'

    Raises:
        ValueError: if the value is an empty iterable (although this not
            applies to string types) or if ``string_func_path`` is ``True`` but
            '::' is not used to define the path to a Python function.
        ModuleNotFoundError: if ``string_func_path`` is ``True`` but the
            provided module can't be imported.

    Returns:
        str: A string resulted from one of the strategies, depends on the input
            data type.
    """
    if isinstance(string, str):
        if not string_func_path:
            return string
        try:
            module_path, resolver_name = string.split("::")
        except ValueError as err:
            if '::' not in string:
                return string
            raise err
        # here raises ``ModuleNotFoundError`` if not module
        mod = importlib.import_module(module_path)
        try:
            resolved = getattr(mod, resolver_name)
        except AttributeError:
            raise ImportError(
                ('Object \'%s\' not found in module \'%s\'') % (
                    resolver_name, module_path))
        return lazy_string(resolved,
                           seed=seed,
                           string_func_path=string_func_path)
    elif isinstance(string, Iterable):
        if not string:
            # Prevent IndexError in ``random.choice``
            raise ValueError(('The iterable used to build a lazy string can'
                              ' not be empty.'))

        if isinstance(string, (set, GeneratorType)):
            string = list(string)
        if seed is not None:
            random.seed(seed)
        return lazy_string(random.choice(string),
                           seed=seed,
                           string_func_path=string_func_path)
    elif isinstance(string, CallableTypes):
        if seed is not None:
            random.seed(seed)
        return lazy_string(string(),
                           seed=seed,
                           string_func_path=string_func_path)
    elif hasattr(string, '__name__'):
        return string.__name__
    return str(string)


def replace_multiple(string, replacements={}):
    """Replaces multiple strings inside a string.

    Args:
        string (str): String to replace.
        replacements (dict): Custom replacements for the string.

    Examples:
        >>> print(replace_multiple('Replace this ".',
        ...                         replacements={'"': 'double quote'}))
        Replace this double quote.

    Returns:
        str: Input string with multiple replaces performed.
    """
    for replacer, replacement in replacements.items():
        string = string.replace(replacer, replacement)
    return string


def escape_single_quote(value):
    '''Escapes single quotes inside a string.

    Args:
        value (str): String for which single quotes will be escaped.

    Raises:
        TypeError: is the value is not a string.

    Examples:
        >>> print(escape_single_quote("Hello I neeed a ' escape."))
        Hello I neeed a \\' escape.
        >>> print(escape_single_quote("Hello I neeed two '' escapes."))
        Hello I neeed two \\'\\' escapes.
        >>> print(escape_single_quote(1))
        Traceback (most recent call last):
          ...
        TypeError: The value '1' can not be escaped because is not a string

    Returns:
        str: The string with single quote characters escaped.
    '''
    try:
        return value.replace("'", "\\'")
    except AttributeError:
        raise TypeError(('The value \'%s\' can not be escaped because is not'
                         ' a string') % str(value))


def escape_double_quote(value):
    '''Escapes double quotes inside a string.

    Args:
        value (str): String for which double quotes will be escaped.

    Raises:
        TypeError: is the value is not a string.

    Examples:
        >>> print(escape_double_quote('Hello I neeed a " escape.'))
        Hello I neeed a \\" escape.
        >>> print(escape_double_quote('Hello I neeed two "" escapes.'))
        Hello I neeed two \\"\\" escapes.
        >>> print(escape_double_quote(1))
        Traceback (most recent call last):
          ...
        TypeError: The value '1' can not be escaped because is not a string

    Returns:
        str: The string with double quote characters escaped.
    '''
    try:
        return value.replace('"', '\\"')
    except AttributeError:
        raise TypeError(('The value \'%s\' can not be escaped because is not'
                         ' a string') % str(value))


def escape_backtick(value):
    '''Escapes backticks inside a string.

    Args:
        value (str): String for which backticks will be escaped.

    Raises:
        TypeError: is the value is not a string.

    Examples:
        >>> print(escape_backtick('Hello I neeed a ` escape.'))
        Hello I neeed a \\` escape.
        >>> print(escape_backtick('Hello I neeed two `` escapes.'))
        Hello I neeed two \\`\\` escapes.
        >>> print(escape_backtick(1))
        Traceback (most recent call last):
          ...
        TypeError: The value '1' can not be escaped because is not a string

    Returns:
        str: The string with double quote characters escaped.
    '''
    try:
        return value.replace('`', '\\`')
    except AttributeError:
        raise TypeError(('The value \'%s\' can not be escaped because is not'
                         ' a string') % str(value))


def lazy_escape_quote_func_by_quote_char(
    quote_char,
    replacers_funcs={'"': escape_double_quote, '\'': escape_single_quote},
    error_msg_schema='Invalid quotation character %(quote_char)s',
    exception_cls=ValueError
):
    '''Factory for quote-escaping functions. It is designed to be used in
    languages utilities implementations to make easier to build quotation
    character functions.

    Arguments:
        quote_char (str): Character to escape.
        replacers_funcs (dict): Mapping of quotation characters to escape
            functions.
        error_msg_schema (str): Schema string that will be the error message
            displayed if the quotation character is not handled by
            ``replacers_funcs`` mapping. Takes the optional named sustituion
            ``'quote_char'`` that will be replaced by the quotation character
            passed in ``quote_char`` parameter.
        exception_cls (type): Exception to be raised if the quotation character
            is not handled by ``replacers_funcs`` mapping.

    Returns:
        str: Quotation character escape function for a string.
    '''
    try:
        return replacers_funcs[quote_char]
    except KeyError:
        if '%(quote_char)s' in error_msg_schema:
            _quote_quote_char = '"' if quote_char == '\'' else '\''
            error_msg_schema = error_msg_schema % {
                'quote_char': (
                    _quote_quote_char + str(quote_char) + _quote_quote_char)
            }
        raise exception_cls(error_msg_schema)



'''Bash curl code snippets generator.'''

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

import json
from collections import OrderedDict

from http_request_codegen.generators.bash._utils import (
    DEFAULT_INDENT,
    DEFAULT_QUOTE_CHAR,
    DEFAULT_WRAP,
    escape_by_quote
)
from http_request_codegen.hrc_valuer import (
    lazy_name_by_parameter,
    lazy_value_by_parameter
)


def _render_options_map(options_map, options_string, url, oneline=False,
                        indent=DEFAULT_INDENT, quote_char=DEFAULT_QUOTE_CHAR):
    response = ''
    if not oneline:
        response += '\\\n'
    for i, (option, value) in enumerate(options_map):
        if value:
            value_string = ' %(quote_char)s%(value)s%(quote_char)s' % {
                'value': escape_by_quote(value, quote_char),
                'quote_char': quote_char
            }
        else:
            value_string = ''
        response += '%(indent)s%(option)s%(value_string)s %(newline)s' % {
            'indent': indent if not oneline else '',
            'option': option,
            'value_string': value_string,
            'newline': '\\\n' if not oneline else ''
        }
    response += '%(indent)s%(url)s' % {
        'indent': indent if not oneline else '',
        'url': url
    }
    return response


def _build_headers(headers, quote_char=DEFAULT_QUOTE_CHAR,
                   content_type='application/x-www-form-urlencoded'):
    map, string = ([], '')
    for name, value in headers.items():
        option = '-H'
        value = ('%(header_name)s: %(header_value)s') % {
            'header_name': name,
            'header_value': escape_by_quote(str(value), quote_char)
        }
        map.append([option, value])

        string += (' %(option)s %(value)s') % {
            'option': option,
            'value': value
        }
        if name.lower() == 'content-type':
            if value in (
                'multipart/form-data',
                'application/json',
                'application/x-www-form-urlencoded',
                'text/plain'
            ):
                content_type = value
    return (map, string, content_type)


def get(url, parameters=[], headers={}, indent=DEFAULT_INDENT,
        quote_char=DEFAULT_QUOTE_CHAR, setup=False, teardown=None,
        oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None, **kwargs):
    '''Pass extra options to 'curl' command in ``kwargs`` parameter. For
    example, to save the response in a file, pass
    ``kwargs={'-o': 'filename.ext'}``:

    ```bash
    curl -o "filename.ext"
    ```
    '''
    '''In this implementation, options values and URLs are not wrapped in
    multiples lines if these values exceed the wrap length.
    '''
    response = ''

    if setup:
        response += str(setup)

    response += 'curl'

    options_string = ''
    options_map = []
    if kwargs or parameters:
        _d_option_included = False
        for option_name, option_value in kwargs.items():
            options_string += ' %(option_name)s' % {'option_name': option_name}
            if option_value:
                option_value_string = escape_by_quote(
                    str(option_value), quote_char)
                options_string += ' ' + option_value_string
            else:
                option_value_string = None
            options_map.append([option_name, option_value_string])

            if option_name == '-X' and option_value != "GET":
                raise ValueError('\'-X\' option, if defined, must be \'GET\'')

            if option_name == '-d':
                _d_option_included = True

        if parameters:
            if _d_option_included:
                raise ValueError(
                    ('You can\'t pass the option \'-d\' and parameters with'
                     ' \'parameters\' value.'))

            parameters_dict = OrderedDict({})
            for parameter in parameters:
                parameter_name = lazy_name_by_parameter(parameter, seed=seed)
                parameter_value = lazy_value_by_parameter(
                    parameter, seed=seed, locale=locale)
                parameters_dict[parameter_name] = parameter_value

            params_string = escape_by_quote(urlencode(parameters_dict),
                                            quote_char)
            options_string += ' -d %(quote_char)s%(params)s%(quote_char)s' % {
                'quote_char': quote_char,
                'params': params_string
            }
            options_map.append(['-d', params_string])

    if headers:
        headers_map, headers_string, _ = _build_headers(headers,
                                                        quote_char=quote_char)
        options_map.extend(headers_map)
        options_string += headers_string

    # 1 here is a space
    if len(options_string) + len(url) + len(response) + 1 < wrap:
        oneline = True
    response += ' '

    response += _render_options_map(options_map,
                                    options_string,
                                    url,
                                    oneline=oneline,
                                    indent=indent,
                                    quote_char=quote_char)

    if teardown:
        response += str(teardown)

    return response


def post(url, parameters=[], files={}, headers={}, indent=DEFAULT_INDENT,
         quote_char=DEFAULT_QUOTE_CHAR, setup=False, teardown=None,
         oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None, **kwargs):
    response = ''

    if setup:
        response += str(setup)

    response += 'curl'

    options_string = ''
    options_map = []

    # Build headers and discover content type
    headers_map, headers_string, content_type = _build_headers(
        headers,
        quote_char=quote_char,
        content_type='application/x-www-form-urlencoded'
                     if not files else 'multipart/form-data')

    _x_post_defined = False
    if kwargs:
        for option_name, option_value in kwargs.items():
            options_string += ' %(option_name)s' % {'option_name': option_name}
            if option_value:
                option_value_str = str(option_value)
                options_string += (' %(quote_char)s%(value)s'
                                   '%(quote_char)s') % {
                    'quote_char': quote_char,
                    'value': option_value_str,
                }
            else:
                option_value_str = None
            options_map.append([option_name, option_value_str])

            if option_name == '-X':
                if option_value != "POST":
                    raise ValueError(
                        '\'-X\' option, if defined, must be \'POST\'')
                else:
                    _x_post_defined = True

    # Add method option
    if not _x_post_defined:
        options_string += ' -X POST'
        options_map.append(['-X', 'POST'])

    # Add parameters
    if parameters:
        parameters_dict = OrderedDict({})
        for parameter in parameters:
            parameter_name = lazy_name_by_parameter(parameter, seed=seed)
            parameter_value = lazy_value_by_parameter(parameter,
                                                      seed=seed,
                                                      locale=locale)
            parameters_dict[parameter_name] = parameter_value

        # parameters codification
        if content_type == 'multipart/form-data':
            for param_name, param_value in parameters_dict.items():
                option_value = '%(name)s=%(value)s' % {
                    'name': param_name,
                    'value': param_value
                }
                options_map.append(['-F', option_value])
                options_string += ' -F %(params)s' % {'params': option_value}
        else:
            encode_func = json.dumps if content_type == 'application/json' \
                else urlencode

            option_value = encode_func(parameters_dict)
            options_map.append(['-d', option_value])
            options_string += ' -d %(quote_char)s%(params)s%(quote_char)s' % {
                'quote_char': quote_char,
                'params': option_value,
            }
    else:
        parameters_dict = {}

    # Add files
    if files:
        for file_param_name, file_data in files.items():
            if file_data is None:
                file_data = lazy_value_by_parameter(
                    {
                        'name': '',
                        'faker': 'faker.providers.file::file_path',
                    },
                    seed=seed,
                    locale=locale)
            elif isinstance(file_data, str):
                file_data = file_data
            else:  # Iterable
                if file_data[0] is None:
                    file_data = lazy_value_by_parameter(
                        {
                            'name': '',
                            'faker': 'faker.providers.file::file_path',
                        },
                        seed=seed,
                        locale=locale)
                else:
                    file_data = file_data[0]
            option_value = '%(name)s=@' % {'name': file_param_name}
            option_value += file_data
            options_string += ' -F %(value)s' % {'value': option_value}
            options_map.append(['-F', option_value])

    # Add headers
    options_string += headers_string
    options_map.extend(headers_map)

    # 1 here is a space
    _current_length = len(options_string) + len(url) + len(response) + 1 \
        + len(headers) * 2 * len(quote_char)
    if _current_length < wrap:
        oneline = True

    # Renderize options
    response += ' '
    response += _render_options_map(options_map,
                                    options_string,
                                    url,
                                    oneline=oneline,
                                    indent=indent,
                                    quote_char=quote_char)

    if teardown:
        response += str(teardown)

    return response



'''Javascript fetch code snippets generator.'''

import os
from collections import OrderedDict


try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from http_request_codegen.generators.javascript._utils import (
    DEFAULT_INDENT,
    DEFAULT_QUOTE_CHAR,
    DEFAULT_WRAP,
    escape_by_quote,
    str_definition
)
from http_request_codegen.hrc_exceptions import (
    raise_post_text_plain_n_parameters_not_1
)
from http_request_codegen.hrc_valuer import (
    lazy_name_by_parameter,
    lazy_value_by_parameter
)


def _promises_chain_render(quote_char=DEFAULT_QUOTE_CHAR,
                           indent=DEFAULT_INDENT,
                           oneline=False):
    return (').then(function(response) {%(newline)s%(indent)s'
            'console.log(response)%(separator)s%(newline)s}).catch('
            'function(error) {%(newline)s%(indent)sconsole.error('
            '%(quote_char)sError:%(quote_char)s, error)%(separator)s'
            '%(newline)s});') % {
        'newline': '\n' if not oneline else '',
        'indent': indent if not oneline else '',
        'quote_char': quote_char,
        'separator': ';' if not oneline else ''
    }


def _headers_render(headers, indent=DEFAULT_INDENT,
                    quote_char=DEFAULT_QUOTE_CHAR, oneline=False,
                    wrap=DEFAULT_WRAP, _comma_at_end=False):
    response = ('%(indent)s%(indent)sheaders:'
                ' {%(newline)s') % {
        'indent': indent if not oneline else '',
        'newline': '\n' if not oneline else '',
    }

    for i, (key, value) in enumerate(headers.items()):
        response += ('%(indent)s%(quote_char)s%(key)s'
                     '%(quote_char)s: %(value)s%(comma)s%(newline)s') % {
            'newline': '\n' if not oneline else '',
            'indent': indent * 3 if not oneline else '',
            'quote_char': quote_char,
            'key': escape_by_quote(key, quote_char),
            'value': str_definition(value,
                                    indent=' ' * (
                                        len(indent) * 3 + len(key) + 4),
                                    quote_char=quote_char,
                                    wrap=wrap),
            'comma': ',' if i < len(headers) - 1 else ''
        }
    response += '%(indent)s}%(comma)s%(newline)s' % {
        'indent': indent * 2 if not oneline else '',
        'comma': ',' if _comma_at_end else '',
        'newline': '\n' if not oneline else ''
    }

    return response


def _kwargs_render(kwargs, indent=DEFAULT_INDENT, wrap=DEFAULT_WRAP,
                   quote_char=DEFAULT_QUOTE_CHAR, oneline=False):
    response = ''
    for i, (key, value) in enumerate(kwargs.items()):
        if isinstance(value, str):
            _value = str_definition(value,
                                    indent=' ' * (
                                        len(indent) * 3 + len(key) + 3),
                                    quote_char=quote_char,
                                    wrap=wrap)
        else:
            _value = str(value)
        response += ('%(indent)s%(quote_char)s%(key)s%(quote_char)s:'
                     ' %(value)s%(comma)s%(newline)s') % {
            'newline': '\n' if not oneline else '',
            'indent': indent * 2 if not oneline else '',
            'quote_char': quote_char,
            'key': escape_by_quote(key, quote_char),
            'value': _value,
            'comma': ',' if i < len(kwargs) - 1 else '',
        }
    return response


def get(url, parameters=[], headers={}, indent=DEFAULT_INDENT,
        quote_char=DEFAULT_QUOTE_CHAR, setup=False, teardown=None,
        oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None,
        **kwargs):
    '''This implementation will emulate browsers\' fetch API by default.
    using Promises-like response processing.

    If you want to simulate a NodeJS environment, pass the parameter ``setup``
    as ``True`` and the next initialization snippet will be prepended to the
    generated code:

    ```javascript
    const fetch = require(\'node-fetch\');
    ```

    For ESM imports emulation, use
    ``setup='import fetch from \\'node-fetch\\';\\n\\n'``, thus the
    initialization snippet will be:

    ```javascript
    import fetch from \'node-fetch\';
    ```

    Of course, you can customize this initialization for other environments.
    For example, polyfill the API using
    [whatwg](https://github.com/github/fetch) as ESM module with
    ``setup='import \\'whatwg-fetch\\';\\n\\n'``:

    ```javascript
    import \'whatwg-fetch\';
    ```
    '''

    '''Implementation details:

    Note that this implementation does not discover the length of the code to
    generate it in multiple lines, like Python requests does. This is because
    in Javascript, promises are chained in multiple lines, thus next structure
    of code would be weird:

    ```javascript
    fetch('<url>').then(function(response) {}, function(error) {});
    ```

    So ``wrap`` argument is used only to wrap strings in multiples lines,
    instead of use it for implement the request in multiples or one line.
    '''

    response = ''

    # initialization
    if setup:
        if isinstance(setup, str):
            response += setup
        else:
            # If `setup == True`, 'node-fetch' for NodeJS is required
            response += ('const fetch = require(%(quote_char)snode-fetch'
                         '%(quote_char)s);%(newline)s%(newline)s') % {
                'newline': '\n' if not oneline else '',
                'quote_char': quote_char
            }

    if parameters:
        parameters_dict = OrderedDict({})
        for parameter in parameters:
            parameters_dict[lazy_name_by_parameter(parameter, seed=seed)] = \
                lazy_value_by_parameter(parameter, seed=seed, locale=locale)
        url = '?'.join([url, urlencode(parameters_dict)])

    if oneline:
        _url = ('%(quote_char)s%(url)s%(quote_char)s') % {
            'quote_char': quote_char,
            'url': escape_by_quote(url, quote_char)
        }
    else:
        _url = str_definition(url, indent=indent, quote_char=quote_char,
                              wrap=wrap)
    response += ('fetch(%(newline)s%(indent)s%(url)s%(comma)s'
                 '%(space)s%(newline)s') % {
        'newline': '\n' if not oneline else '',
        'indent': indent if not oneline else '',
        'url': _url,
        'comma': ',' if (parameters or headers or kwargs) else '',
        'space': ' ' if (oneline and (parameters or headers or kwargs))
                 else '',
    }

    # options render
    if headers or kwargs:
        response += '%(indent)s{%(newline)s' % {
            'indent': indent if not oneline else '',
            'newline': '\n' if not oneline else ''
        }

    if headers:
        response += _headers_render(headers, indent=indent, oneline=oneline,
                                    quote_char=quote_char, wrap=wrap,
                                    _comma_at_end=bool(kwargs))

    if kwargs:
        response += _kwargs_render(kwargs, indent=indent, oneline=oneline,
                                   quote_char=quote_char, wrap=wrap)

    if headers or kwargs:
        response += '%(indent)s}%(newline)s' % {
            'indent': indent if not oneline else '',
            'newline': '\n' if not oneline else '',
        }

    response += _promises_chain_render(quote_char=quote_char, indent=indent,
                                       oneline=oneline)

    if teardown:
        response += teardown

    return response


def post(url, parameters=[], files={}, headers={}, indent=DEFAULT_INDENT,
         quote_char=DEFAULT_QUOTE_CHAR, setup=True, teardown=None,
         oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None,
         **kwargs):
    # (no setup -> web / setup -> node)
    response = ''

    # Discover content-type
    content_type = 'application/x-www-form-urlencoded'
    for key, value in headers.items():
        if key.lower() == 'content-type':
            content_type = value
            break
    if content_type.startswith('multipart/form-data') or files:
        content_type = 'multipart/form-data'

    # Initialization will depend on content type
    if setup:
        if isinstance(setup, str):
            response += setup
        else:
            if content_type == 'multipart/form-data':
                response += ('const fs = require(%(quote_char)s'
                             'fs%(quote_char)s);%(newline)s%(newline)s') % {
                    'quote_char': quote_char,
                    'newline': '\n' if not oneline else '',
                }
            response += ('const fetch = require(%(quote_char)s'
                         'node-fetch%(quote_char)s);') % {
                'quote_char': quote_char,
            }
            if content_type == 'multipart/form-data':
                response += ('%(newline)sconst FormData = require('
                             '%(quote_char)sform-data%(quote_char)s);') % {
                    'quote_char': quote_char,
                    'newline': '\n' if not oneline else '',
                }
            if not oneline:
                response += '\n\n'

    if content_type == 'multipart/form-data':
        body = 'formData'

        # if we are sending files from the browser, select all files
        if not setup and files:
            response += ('const files = document.querySelector('
                         '%(quote_char)sinput[type=%(other_quote_char)s'
                         'file%(other_quote_char)s]%(quote_char)s);'
                         '%(newline)s%(newline)s') % {
                'quote_char': quote_char,
                'other_quote_char': '"' if quote_char == '\'' else '\'',
                'newline': '\n' if not oneline else ''
            }

        response += 'const formData = new FormData();%(newline)s' % {
            'newline': '\n' if not oneline else '',
        }

        # parameters render
        for parameter in parameters:
            name = lazy_name_by_parameter(parameter, seed=seed)
            value = lazy_value_by_parameter(
                parameter, seed=seed, locale=locale)

            # 20 here is the length of `formData.append(, );`
            _multiline_param = False
            if 20 + len(quote_char) * 4 + len(name) + len(value) > wrap:
                _multiline_param = True

            if _multiline_param:
                _param_name = str_definition(name, indent=indent,
                                             quote_char=quote_char, wrap=wrap)
                _param_value = str_definition(value, indent=indent,
                                              quote_char=quote_char, wrap=wrap)
            else:
                _value_name_schema = '%(quote_char)s%(string)s%(quote_char)s'
                _param_name = _value_name_schema % {
                    'quote_char': quote_char,
                    'string': escape_by_quote(name, quote_char),
                }
                _param_value = _value_name_schema % {
                    'quote_char': quote_char,
                    'string': escape_by_quote(value, quote_char),
                }

            response += ('formData.append(%(newline)s%(indent)s'
                         '%(param_name)s,%(space)s'
                         '%(newline)s%(indent)s%(param_value)s'
                         '%(newline)s);') % {
                'newline': '\n' if (not oneline and _multiline_param) else '',
                'indent': indent if (not oneline and _multiline_param) else '',
                'space': ' ' if (oneline or _multiline_param) else '',
                'param_name': _param_name,
                'param_value': _param_value,
            }
            if not oneline:
                response += '\n'

        # files render
        for i, (file_param_name, file_data) in enumerate(files.items()):
            response += 'formData.append('
            if not oneline:
                _file_param_name = str_definition(file_param_name,
                                                  quote_char=quote_char,
                                                  wrap=wrap,
                                                  indent=indent)
            else:
                _file_param_name = ('%(quote_char)s'
                                    '%(file_param_name)s'
                                    '%(quote_char)s') % {
                    'quote_char': quote_char,
                    'file_param_name': escape_by_quote(file_param_name,
                                                       quote_char)
                }

            response += ('%(newline)s%(indent)s%(file_param_name)s,%(space)s'
                         '%(newline)s%(indent)s') % {
                'newline': '\n' if not oneline else '',
                'indent': indent if not oneline else '',
                'file_param_name': _file_param_name,
                'space': ' ' if oneline else '',
            }

            if setup:
                response += 'fs.createReadStream('  # length: 20

                if isinstance(file_data, str):
                    filepath = file_data
                else:
                    filepath = file_data[0] if file_data is not None else None
                if filepath is None:
                    filepath = lazy_value_by_parameter(
                        {
                            'name': '',
                            'faker': 'faker.providers.file::file_path',
                        },
                        seed=seed,
                        locale=locale)

                if not oneline:
                    _filepath = str_definition(
                        filepath,
                        quote_char=quote_char,
                        wrap=wrap,
                        indent=' ' * (len(indent) + 20))
                    _filename = str_definition(
                        os.path.basename(filepath),
                        quote_char=quote_char,
                        wrap=wrap,
                        indent=' ' * (len(indent) * 2 + 10))  # 'filename: '
                else:
                    _filepath = '%(quote_char)s%(filepath)s%(quote_char)s' % {
                        'quote_char': quote_char,
                        'filepath': escape_by_quote(filepath, quote_char)
                    }
                    _filename = '%(quote_char)s%(filename)s%(quote_char)s' % {
                        'quote_char': quote_char,
                        'filename': escape_by_quote(
                            os.path.basename(filepath), quote_char)
                    }

                response += ('%(filepath)s),%(space)s%(newline)s%(indent)s{'
                             '%(newline)s%(indent)s%(indent)s'
                             'filename: %(filename)s') % {
                    'space': ' ' if oneline else '',
                    'newline': '\n' if not oneline else '',
                    'indent': indent if not oneline else '',
                    'filepath': _filepath,
                    'filename': _filename
                }
                if not isinstance(file_data, str) and file_data is not None:
                    if len(file_data) > 1:
                        if not oneline:
                            _content_type = str_definition(
                                file_data[1],
                                quote_char=quote_char,
                                wrap=wrap,
                                indent=' ' * (len(indent) * 2 + 13))
                        else:
                            _content_type = ('%(quote_char)s%(content_type)s'
                                             '%(quote_char)s') % {
                                'quote_char': quote_char,
                                'content_type': escape_by_quote(
                                    file_data[1], quote_char)
                            }

                        response += (',%(newline)s%(indent)s%(indent)s'
                                     'contentType: %(content_type)s') % {
                            'content_type': _content_type,
                            'newline': '\n' if not oneline else '',
                            'indent': indent if not oneline else ''
                        }
                response += '%(newline)s%(indent)s}%(newline)s' % {
                    'newline': '\n' if not oneline else '',
                    'indent': indent if not oneline else ''
                }
            else:
                # TODO: Manage content_type and filename?
                response += 'inputs[%(input_index)d].files[0]%(newline)s' % {
                    'input_index': i,
                    'newline': '\n' if not oneline else ''
                }
            response += ');%(newline)s' % {
                'newline': '\n' if not oneline else ''
            }
        response += '\n'
    elif content_type == 'text/plain':
        if len(parameters) != 1:
            raise_post_text_plain_n_parameters_not_1(len(parameters))

        _value = lazy_value_by_parameter(parameters[0],
                                         seed=seed,
                                         locale=locale)
        body = str_definition(_value,
                              indent=' ' * (len(indent) * 2 + 6),
                              quote_char=quote_char,
                              wrap=wrap)
    else:  # 'application/json' or 'application/x-www-form-urlencoded'
        object_content = ''

        for i, parameter in enumerate(parameters):
            name = lazy_name_by_parameter(parameter, seed=seed)
            value = lazy_value_by_parameter(
                parameter, seed=seed, locale=locale)

            if oneline:
                name_def = '%(quote_char)s%(name)s%(quote_char)s' % {
                    'name': escape_by_quote(name, quote_char),
                    'quote_char': quote_char
                }
                value_def = '%(quote_char)s%(value)s%(quote_char)s' % {
                    'value': escape_by_quote(value, quote_char),
                    'quote_char': quote_char
                }
            else:
                name_def = str_definition(name,
                                          indent=indent * 3,
                                          quote_char=quote_char,
                                          wrap=wrap)
                _value_def_indent = ' ' * (
                    len(indent) * 3 + len(name_def) + 2)
                value_def = str_definition(value,
                                           indent=_value_def_indent,
                                           quote_char=quote_char,
                                           wrap=wrap)

            object_content += ('%(indent)s%(name)s: %(value)s%(comma)s'
                               '%(newline)s') % {
                'name': name_def,
                'indent': indent * 3,
                'value': value_def,
                'comma': ',' if i < (len(parameters) - 1) else '',
                'newline': '\n' if (
                    not oneline and i < (len(parameters) - 1)) else ''
            }

        if object_content:
            body = ('%(object_wrapper)s({%(newline)s'
                    '%(object_content)s%(newline)s%(indent)s})') % {
                'object_wrapper': 'JSON.stringify'
                if content_type == 'application/json'
                else 'new URLSearchParams',
                'object_content': object_content,
                'newline': '\n' if not oneline else '',
                'indent': indent * 2,
            }
        else:
            body = ''

    response += ('fetch(%(newline)s%(indent)s%(url)s,%(space)s%(newline)s'
                 '%(indent)s{%(indent)s%(newline)s%(indent)s%(indent)smethod:'
                 ' %(quote_char)sPOST%(quote_char)s%(comma)s%(newline)s') % {
        'newline': '\n' if not oneline else '',
        'indent': indent if not oneline else '',
        'space': ' ' if oneline else '',
        'quote_char': quote_char,
        'comma': ',' if (body or kwargs or headers) else '',
        'url': str_definition(url, quote_char=quote_char,
                              indent=indent, wrap=wrap),
    }

    if body:
        response += ('%(indent)s%(indent)sbody:'
                     ' %(body)s%(comma)s%(newline)s') % {
            'newline': '\n' if not oneline else '',
            'indent': indent if not oneline else '',
            'comma': ',' if (kwargs or headers) else '',
            'body': body,
        }

    # headers render
    if headers:
        response += _headers_render(headers, indent=indent, oneline=oneline,
                                    quote_char=quote_char, wrap=wrap,
                                    _comma_at_end=bool(kwargs))

    # kwargs render
    if kwargs:
        response += _kwargs_render(kwargs, indent=indent, oneline=oneline,
                                   quote_char=quote_char, wrap=wrap)

    response += '%(indent)s}%(newline)s' % {
        'indent': indent if not oneline else '',
        'newline': '\n' if not oneline else ''
    }

    response += _promises_chain_render(quote_char=quote_char, indent=indent,
                                       oneline=oneline)

    if teardown:
        response += teardown

    return response



'''Utilities for Javascript HTTP request generators.'''

from http_request_codegen.hrc_string import (
    escape_backtick,
    escape_double_quote,
    escape_single_quote,
    lazy_escape_quote_func_by_quote_char
)


DEFAULT_INDENT = '  '
DEFAULT_QUOTE_CHAR = "'"
DEFAULT_WRAP = 80


def escape_by_quote(string, char):
    '''Escapes the quote characters of a string.

    Args:
        string (str): String whose qutation characters must be escaped.
        char (str): Quotation character to escape.

    Raises:
        ValueError: If the character to escape is an invalid Javascript string
            quotation character.
        TypeError: If the value to escape is not a string.

    Examples:
        >>> print(escape_by_quote("I need single quote ' escape.", "'"))
        I need single quote \\' escape.

        >>> print(escape_by_quote("I need two single quote '' escapes.", "'"))
        I need two single quote \\'\\' escapes.

        >>> print(escape_by_quote('I need double quote " escape.', '"'))
        I need double quote \\" escape.

        >>> print(escape_by_quote('I need two double quote "" escapes.', '"'))
        I need two double quote \\"\\" escapes.

        >>> escape_by_quote(12, '"')
        Traceback (most recent call last):
          ...
        TypeError: The value '12' can not be escaped because is not a string

        >>> escape_by_quote('I need escape of ? character.', '?')
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Javascript quotation character

        >>> escape_by_quote(12, '?')
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Javascript quotation character

    Returns:
        str: The original string with the specified character escaped.
    '''
    return lazy_escape_quote_func_by_quote_char(
        char,
        replacers_funcs={
            '"': escape_double_quote,
            '\'': escape_single_quote,
            '`': escape_backtick
        },
        error_msg_schema=('%(quote_char)s is an invalid Javascript quotation'
                          ' character')
    )(string)


def str_definition(string, indent=DEFAULT_INDENT,
                   quote_char=DEFAULT_QUOTE_CHAR, wrap=DEFAULT_WRAP,
                   _escape=True):
    '''Creates a definition of a Javascript string, multilining it if
    neccesary. Does not handle spaces at all wrapping it, so it's useful to
    efficiently wrap non-spaced strings like URLs.

    Args:
        string (str): String to be reproducted.
        indent (str): Indentation string. This defines the space at the left of
            the string in the code with respect to column 0 of the code.
        quote_char (str): Quotation mark character used in the string. Must be
            a valid Python string quotation character or a `ValueError` will be
            raised.
        wrap (int): Maximum anchor in which the string will be wrapped in
            multiples lines with respect to column 0 of the code.
        _escape (bool): Defines if the reproducted string must be escaped by
            the given ``quote_char`` argument.

    Raises:
        ValueError: if the argument ``quote_char`` is not a valid Javascript
            string quotation character.

    Examples:

        >>> print(str_definition(
        ...           'Lorem Ipsum es simplemente el texto de'
        ...           + ' relleno de las imprentas y archivos de texto. Lorem'
        ...           + ' Ipsum ha sido el texto de relleno estándar de las'
        ...           + ' industrias desde el año 1500.',
        ...           wrap=70, indent=''))
        'Lorem Ipsum es simplemente el texto de relleno de las imprentas y a'
        + 'rchivos de texto. Lorem Ipsum ha sido el texto de relleno estánda'
        + 'r de las industrias desde el año 1500.'

    Returns:
        str: Javascript string definition reproducted.
    '''
    string_escaped = string if not _escape else \
        escape_by_quote(string, quote_char)

    if len(string) + len(indent) + len(quote_char) * 2 < wrap:
        return '%(quote_char)s%(value)s%(quote_char)s' % {
            'quote_char': quote_char,
            'value': string_escaped,
        }

    indent_length = len(indent)
    response = '%(quote_char)s' % {
        'quote_char': quote_char
    }
    _chars_in_current_line = len(response) + indent_length
    for i, ch in enumerate(string):
        response += ch
        _chars_in_current_line += 1
        if _chars_in_current_line >= wrap - 2:
            if i >= len(string) - 1:
                break
            response += '%(quote_char)s\n%(indent)s+ %(quote_char)s' % {
                'quote_char': quote_char,
                'indent': indent,
            }
            _chars_in_current_line = 3 + indent_length  # ' ('
    response += '%(quote_char)s%(newline)s%(indent)s' % {
        'newline': '\n' if _chars_in_current_line >= wrap - 1 else '',
        'indent': indent if _chars_in_current_line >= wrap - 1 else '',
        'quote_char': quote_char,
    }
    return response



'''Python requests code snippets generator.'''

from collections import OrderedDict

from http_request_codegen.generators.python._utils import (
    DEFAULT_INDENT,
    DEFAULT_QUOTE_CHAR,
    DEFAULT_WRAP,
    dict_definition,
    escape_by_quote,
    kwarg_definition,
    kwarg_definition_dict_valued,
    str_definition
)
from http_request_codegen.hrc_exceptions import (
    raise_post_text_plain_n_parameters_not_1
)
from http_request_codegen.hrc_valuer import (
    lazy_name_by_parameter,
    lazy_value_by_parameter
)


def get(url, parameters=[], headers={}, indent=DEFAULT_INDENT,
        quote_char=DEFAULT_QUOTE_CHAR, setup=True, teardown=None,
        oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None,
        **kwargs):
    '''Parameters are passed using
    [``requests.get``](https://requests.readthedocs.io/en/api/#requests.get)
    function ``params`` parameter, not by appending ``?foo=bar&...`` to the
    URL. If you want this behaviour, build the passed URL using
    [``lazy_name_by_parameter``](/reference#lazy_name_by_parameter) and
    [``lazy_value_by_parameter``](/reference#lazy_value_by_parameter) functions
    instead of use the ``parameters`` argument.

    If you want to import more modules in the initialization snippet, keep
    in mind that you must provide ``import requests`` line also in the
    ``setup`` argument. For example,
    ``setup=\'import requests\\nimport foo\\n\\n\'`` will render as:

    ```python
    import requests
    import foo

    requests.get('<url>'...
    ```
    '''
    _oneline = oneline
    response = ''

    # initialization
    setup_length = 0
    if setup:
        if isinstance(setup, str):
            response += setup
        else:
            response += ('import requests%(separator)s'
                         '%(newline)s%(newline)s') % {
                'separator': ';' if oneline else '',
                'newline': '\n' if not oneline else '',
            }
        if oneline:
            setup_length += len(response)

    # url length
    url_length = len(url) + 21  # 'req = requests.get(' (19) + '' (2)

    # parameters length
    parameters_line_length = 0
    if parameters:
        parameters_keys = OrderedDict({})
        parameters_line_length = 9  # 'params={}'
        for parameter in parameters:
            name = escape_by_quote(
                lazy_name_by_parameter(parameter, seed=seed),
                quote_char
            )
            value = str_definition(
                lazy_value_by_parameter(parameter, seed=seed, locale=locale),
                quote_char=quote_char,
                indent=indent + (' ' * (8 + len(name))),
                wrap=wrap
            )
            parameters_keys[name] = value
            parameters_line_length += len(name) + 4 + len(value)
        parameters_line_length += len(parameters) - 1  # commas except last
        if headers or kwargs:
            parameters_line_length += 2  # ', '

    headers_line_length = 0
    if headers:
        headers_keys = {} if not isinstance(headers, OrderedDict) \
            else OrderedDict({})
        headers_line_length += 10  # headers={}
        for key, value in headers.items():
            escaped_key = escape_by_quote(key, quote_char)
            escaped_value = escape_by_quote(value, quote_char)
            headers_keys[escaped_key] = escaped_value
            # 2 is ': '
            headers_line_length += \
                len(quote_char) * 4 + 2 + len(escaped_key) + len(escaped_value)
        headers_line_length += len(headers) - 1  # commas except last
        if kwargs:
            headers_line_length += 2  # ', '

    kwargs_line_length = 0
    if kwargs:
        kwargs_line_length = 2
        for key, value in kwargs.items():
            kwargs_line_length += len(key) + len(str(value))
            if isinstance(value, str):
                kwargs_line_length += 2
        kwargs_line_length += len(kwargs) - 1  # commas except last

    # oneliner by wrapping
    #   here + 1 is function call end ')' and - 1 is the character wrapping,
    #   so both are negated
    if setup_length + url_length + parameters_line_length + \
            headers_line_length + kwargs_line_length < wrap:
        oneline = True

    # url
    response += ('req = requests.get(%(newline)s%(indent)s%(url)s%(newline2)s'
                 '%(comma)s%(space)s%(newline3)s') % {
        'url': ('%(quote_char)s%(url)s%(quote_char)s' % {
            'url': url,
            'quote_char': quote_char
        }) if oneline else str_definition(url,
                                          indent=indent,
                                          quote_char=quote_char,
                                          wrap=wrap),
        'indent': indent if not oneline else '',
        'newline': '\n' if not oneline else '',
        'newline2': '\n' if not oneline and (
            not headers and not parameters and not kwargs) else '',
        'comma': ',' if headers or parameters or kwargs else '',
        'space': ' ' if oneline and (parameters or headers or kwargs) else '',
        'newline3': '\n' if not oneline and (
            headers or parameters or kwargs) else ''
    }

    # parameters render
    if parameters:
        response += '%(indent)sparams={%(newline)s' % {
            'indent': indent if not oneline else '',
            'newline': '\n' if not oneline else '',
        }
        for i, (pname, pvalue) in enumerate(parameters_keys.items()):
            response += ('%(indent)s%(quote_char)s%(parameter_name)s'
                         '%(quote_char)s: %(value)s%(comma)s') % {
                'parameter_name': pname,
                'indent': indent * 2 if not oneline else '',
                'quote_char': quote_char,
                'value': pvalue,
                'comma': ',' if i < len(parameters) - 1 else '',
            }
            response += '\n' if not oneline else ''

        response += '%(indent)s}%(comma)s%(newline)s' % {
            'indent': indent if not oneline else '',
            'comma': ',' if (headers or kwargs) else '',
            'newline': '\n' if not oneline else ('' if (
                not headers and not kwargs) else ' '),
        }

    # headers render
    if headers:
        response += '%(header)s%(comma)s%(newline)s' % {
            'header': kwarg_definition_dict_valued(
                'headers', headers_keys,
                indent=indent if not oneline else '',
                quote_char=quote_char,
                newline='\n' if not oneline else '',
                _escape_keys=False, _escape_values=False),
            'newline': '\n' if not oneline else (
                '' if not kwargs else ' '),
            'comma': ',' if kwargs else '',
        }

    # kwargs render
    if kwargs:
        for i, (key, value) in enumerate(kwargs.items()):
            response += '%(indent)s%(repr_kwarg)s%(comma)s%(newline)s' % {
                'indent': '' if oneline or isinstance(value, dict) else indent,
                'repr_kwarg': kwarg_definition(
                    key,
                    value,
                    indent='' if oneline else indent,
                    quote_char=quote_char,
                    wrap=wrap),
                'comma': ',' if i < len(kwargs) - 1 else '',
                'newline': '\n' if not oneline else (
                    ' ' if i < len(kwargs) - 1 else ''),
            }

    response += ')%(separator)s' % {'separator': ';' if _oneline else ''}

    if teardown:
        response += str(teardown)
    return response


def post(url, parameters=[], files={}, headers={}, indent=DEFAULT_INDENT,
         quote_char=DEFAULT_QUOTE_CHAR, setup=True, teardown=None,
         oneline=False, wrap=DEFAULT_WRAP, seed=None, locale=None,
         **kwargs):
    '''POST method code generator for Python requests library.'''
    # There are 4 possibilities of arguments build since we allow 4 forms
    #   of doing POST requests, depends on 'Content-Type' header content
    #   and the definition of 'files' optional argument. As default,
    #   'application/x-www-form-urlencoded'
    #   - len(files) > 0 -> files={}
    #   - Content-Type: 'multipart/form-data; boundary=-----...' -> files={}
    #   - Content-Type: 'application/x-www-form-urlencoded' -> data={}
    #   - Content-Type: 'text/plain' -> data=''
    #   - Content-Type: 'application/json' -> json={}
    _oneline = oneline
    response = ''

    # initialization
    setup_length = 0
    if setup:
        if isinstance(setup, str):
            response += setup
        else:
            response += ('import requests%(separator)s'
                         '%(newline)s%(newline)s') % {
                'separator': ';' if oneline else '',
                'newline': '\n' if not oneline else '',
            }
        if oneline:
            setup_length += len(response)

    # url length
    url_length = len(url) + 22  # 'req = requests.post(' (20) + '' (2)

    # headers length and content-type discovering
    if files:
        content_type = 'multipart/form-data'
        _content_type_found = True
    else:
        content_type = 'application/x-www-form-urlencoded'
        _content_type_found = False

    headers_line_length = 0
    if headers:
        headers_keys = {} if not isinstance(headers, OrderedDict) \
            else OrderedDict({})
        headers_line_length += 10  # headers={}
        for key, value in headers.items():
            escaped_key = escape_by_quote(key, quote_char)
            escaped_value = escape_by_quote(value, quote_char)
            headers_keys[escaped_key] = escaped_value
            # 2 is ': '
            headers_line_length += \
                len(quote_char) * 4 + 2 + len(escaped_key) + len(escaped_value)
            if not _content_type_found:
                _key_lower = str(key).lower()
                if _key_lower == 'content-type':
                    if 'text/plain' in value:
                        content_type = 'text/plain'
                        _content_type_found = True
                    elif 'application/json' in value:
                        content_type = 'application/json'
                        _content_type_found = True
        headers_line_length += len(headers) - 1  # commas except last
        if kwargs:
            headers_line_length += 2  # ', '

    if content_type == 'text/plain' and len(parameters) != 1:
        raise_post_text_plain_n_parameters_not_1(len(parameters))

    # data/json length
    parameters_line_length = 0
    if parameters:
        parameters_keys = OrderedDict({})

        # length depends on kwarg used in the request
        parameters_line_length = {
            'multipart/form-data': 0,                # files={}
            'application/json': 7,                   # json={}
            'application/x-www-form-urlencoded': 7,  # data={}
        }.get(content_type, 5)  # (text/*)             data=

        for parameter in parameters:
            if content_type == 'text/plain':
                name = ''
            else:
                name = escape_by_quote(
                    lazy_name_by_parameter(parameter, seed=seed),
                    quote_char
                )

            # JSON must accepts other data types than string
            _param_value = parameter.get('value')
            if content_type == 'application/json' and isinstance(
                    _param_value, (int, float, bool)):
                value = _param_value
                parameters_line_length += len(name) + 4 + len(str(value))
            else:
                if content_type == 'text/plain':
                    _param_value_def_indent = indent + (' ' * 5)
                else:
                    _param_value_def_indent = indent + (' ' * (8 + len(name)))
                value = str_definition(
                    lazy_value_by_parameter(parameter,
                                            seed=seed,
                                            locale=locale),
                    quote_char=quote_char,
                    indent=_param_value_def_indent,
                    wrap=wrap
                )
                parameters_line_length += len(name) + 4 + len(value)

            if content_type == 'text/plain':
                parameters_line_length += len(value) + 2
            elif content_type != 'application/json':
                parameters_line_length += len(name) + 4 + len(value)
            parameters_keys[name] = value
        parameters_line_length += len(parameters) - 1  # commas except last
        if files or headers or kwargs:
            parameters_line_length += 2  # ', '

    # files length
    files_line_length = 0
    if files:
        files_line_length = 8  # files={}
        files_keys = {} if not isinstance(files, OrderedDict) \
            else OrderedDict({})
        for key, value in files.items():
            escaped_key = escape_by_quote(key, quote_char)
            files_line_length += len(escaped_key) + 4

            files_keys[escaped_key] = []

            if isinstance(value, str) or value is None:
                # random filepath
                if value is None:
                    value = lazy_value_by_parameter(
                        {
                            'name': '',
                            'faker': 'faker.providers.file::file_path',
                        },
                        seed=seed,
                        locale=locale)

                escaped_value = escape_by_quote(value, quote_char)
                files_keys[escaped_key].append(escaped_value)

            else:  # iterable
                if len(value) > 2:
                    # file headers
                    _fheaders = {}
                    for hkey, hvalue in value[2].items():
                        _escaped_key = escape_by_quote(hkey, quote_char)
                        files_line_length += len(_escaped_key) + 4

                        _escaped_value = escape_by_quote(hvalue, quote_char)
                        files_line_length += len(_escaped_value) + 2

                        _fheaders[_escaped_key] = _escaped_value
                    # commas except last + prev ', '
                    files_line_length += len(value[2]) + 2
                    files_keys[escaped_key].append(_fheaders)
                if len(value) > 1:
                    # file content type
                    _escaped_value = escape_by_quote(value[1], quote_char)
                    files_line_length += len(_escaped_value) + 4  # prev ', '
                    files_keys[escaped_key].insert(0, _escaped_value)

                # random filepath
                if value[0] is None:
                    value = list(value)
                    value[0] = lazy_value_by_parameter(
                        {
                            'name': '',
                            'faker': 'faker.providers.file::file_path',
                        },
                        seed=seed,
                        locale=locale)
                escaped_value = escape_by_quote(value[0], quote_char)
                files_keys[escaped_key].insert(0, escaped_value)

            _filepath = value if isinstance(value, str) else value[0]
            _indent = indent * 4
            _filepath_def = str_definition(
                _filepath,
                indent=_indent,
                quote_char=quote_char,
                wrap=wrap)
            _open_file_multiline = len(_filepath_def) + len(_indent) > wrap
            _open_file_string = ('open(%(newline)s%(indent1)s%(value)s,'
                                 '%(newline)s%(indent2)s%(quote_char)srb'
                                 '%(quote_char)s%(newline)s%(indent3)s)') % {
                'quote_char': quote_char,
                'value': _filepath_def,
                'newline': '\n' if _open_file_multiline else '',
                'indent1': _indent if _open_file_multiline else '',
                'indent2': _indent if _open_file_multiline else ' ',
                'indent3': indent * 3 if _open_file_multiline else '',
            }
            files_keys[escaped_key].insert(1, _open_file_string)

            # 2 filename quotes + () of tuple + , open('', 'rb')
            files_line_length += len(escaped_value) * 2 + 20

        files_line_length += len(parameters) - 1  # commas except last

    kwargs_line_length = 0
    if kwargs:
        kwargs_line_length = 2
        for key, value in kwargs.items():
            kwargs_line_length += len(key) + len(str(value))
            if isinstance(value, str):
                kwargs_line_length += 2
        kwargs_line_length += len(kwargs) - 1  # commas except last

    # oneliner by wrapping
    #   here + 1 is function call end ')' and - 1 is the character wrapping,
    #   so both are negated
    if setup_length + url_length + parameters_line_length + \
            headers_line_length + files_line_length + \
            kwargs_line_length < wrap:
        oneline = True

    # url
    response += ('req = requests.post(%(newline)s%(indent)s%(url)s%(newline2)s'
                 '%(comma)s%(space)s%(newline3)s') % {
        'url': ('%(quote_char)s%(url)s%(quote_char)s' % {
            'url': url, 'quote_char': quote_char
        }) if oneline else str_definition(url,
                                          indent=indent,
                                          quote_char=quote_char,
                                          wrap=wrap),
        'indent': indent if not oneline else '',
        'newline': '\n' if not oneline else '',
        'newline2': '\n' if not oneline and (
            not headers and not files and not parameters and not kwargs
        ) else '',
        'comma': ',' if (files or headers or parameters or kwargs) else '',
        'space': ' ' if oneline and (
            files or parameters or headers or kwargs) else '',
        'newline3': '\n' if not oneline and (
            files or parameters or headers or kwargs) else ''
    }

    # parameters render
    if parameters:
        response += '%(indent)s%(kwarg_name)s=' % {
            'indent': indent if not oneline else '',
            'kwarg_name': 'json' if content_type == 'application/json'
            else 'data'
        }
        if content_type == 'text/plain':
            response += '%(value)s%(comma)s%(newline)s' % {
                'value': parameters_keys[''],
                'comma': ',' if (
                    not oneline or files or headers or kwargs) else '',
                'newline': '\n' if not oneline else ' '
            }
        else:
            response += '{%(newline)s' % {
                'newline': '\n' if not oneline else '',
            }
            for i, (pname, pvalue) in enumerate(parameters_keys.items()):
                response += ('%(indent)s%(quote_char)s%(parameter_name)s'
                             '%(quote_char)s: %(value)s%(comma)s') % {
                    'parameter_name': pname,
                    'indent': indent * 2 if not oneline else '',
                    'quote_char': quote_char,
                    'value': pvalue,
                    'comma': ',' if i < len(parameters) - 1 else '',
                }
                response += '\n' if not oneline else ''

            response += '%(indent)s}%(comma)s%(newline)s' % {
                'indent': indent if not oneline else '',
                'comma': ',' if (files or headers or kwargs) else '',
                'newline': '\n' if not oneline else ('' if (
                    not headers and not files and not kwargs) else ' '),
            }

    # files render
    if files:
        response += '%(indent)sfiles={%(newline)s' % {
            'indent': indent if not oneline else '',
            'newline': '\n' if not oneline else '',
        }

        for i, (fkey, fvalue) in enumerate(files_keys.items()):
            _value = ('(%(newline1)s%(indent)s%(fpath)s,%(newline2)s'
                      '%(indent)s%(open_file_string)s') % {
                'newline1': '\n' if not oneline else '',
                'newline2': '\n' if not oneline else ' ',
                'indent': indent * 3 if not oneline else '',
                'fpath': str_definition(
                    fvalue[0],
                    quote_char=quote_char,
                    indent=indent * 3,
                    wrap=wrap),
                'open_file_string': fvalue[1]
            }
            if len(fvalue) > 2:
                _value += ',%(newline)s%(indent)s%(content_type)s' % {
                    'content_type': str_definition(
                        fvalue[2],
                        quote_char=quote_char,
                        indent=indent * 3,
                        wrap=wrap),
                    'newline': '\n' if not oneline else '',
                    'indent': indent * 3 if not oneline else ' '
                }
            if len(fvalue) > 3:
                _value += ',%(newline)s%(indent)s%(headers)s' % {
                    'headers': dict_definition(
                        fvalue[3],
                        indent=indent if not oneline else '',
                        indent_depth=3,
                        quote_char=quote_char,
                        newline='\n' if not oneline else '',
                        _escape_keys=False, _escape_values=False),
                    'newline': '\n' if not oneline else '',
                    'indent': indent * 3 if not oneline else ' '
                }
            _value += '%(newline)s%(indent)s)' % {
                'newline': '' if oneline else '\n',
                'indent': '' if oneline else indent * 2
            }
            response += ('%(indent)s%(quote_char)s%(key)s'
                         '%(quote_char)s: %(value)s%(comma)s') % {
                'key': fkey,
                'indent': indent * 2 if not oneline else '',
                'quote_char': quote_char,
                'value': _value,
                'comma': ',' if i < len(files_keys) - 1 else '',
            }
            response += '\n' if not oneline else ''

        response += '%(indent)s}%(comma)s%(newline)s' % {
            'indent': indent if not oneline else '',
            'comma': ',' if (headers or kwargs) else '',
            'newline': '\n' if not oneline else (' ' if (
                headers or kwargs) else ''),
        }

    if headers:
        response += '%(headers)s%(comma)s%(newline)s' % {
            'headers': kwarg_definition_dict_valued(
                'headers', headers_keys,
                indent=indent if not oneline else '',
                quote_char=quote_char,
                newline='\n' if not oneline else '',
                _escape_keys=False, _escape_values=False),
            'newline': '\n' if not oneline else (
                '' if not kwargs else ' '),
            'comma': ',' if kwargs else '',
        }

    # kwargs render
    if kwargs:
        for i, (key, value) in enumerate(kwargs.items()):
            response += '%(indent)s%(repr_kwarg)s%(comma)s%(newline)s' % {
                'indent': '' if oneline or isinstance(value, dict) else indent,
                'repr_kwarg': kwarg_definition(
                    key,
                    value,
                    indent='' if oneline else indent,
                    indent_depth=1,
                    quote_char=quote_char,
                    wrap=wrap),
                'comma': ',' if i < len(kwargs) - 1 else '',
                'newline': '\n' if not oneline else (
                    ' ' if i < len(kwargs) - 1 else ''),
            }

    response += ')%(separator)s' % {'separator': ';' if _oneline else ''}

    if teardown:
        response += str(teardown)
    return response



'''http-request-codegen public API.'''

from http_request_codegen.hrc_factory import (
    DEFAULT_LANGUAGE,
    get_func_by_lang_impl_method
)
from http_request_codegen.hrc_string import lazy_string


def generate_http_request_code(language=None, impl=None, method='GET',
                               url='http://localhost', parameters=[],
                               headers={}, files={}, indent=None,
                               quote_char='\'', setup=None, teardown=None,
                               oneline=False, seed=None, locale=None, wrap=80,
                               **kwargs):
    '''Generates a code snippet of an HTTP request for a library of a given
    programming language or a CLI of a program, based on a valid HTTP method
    and a specification of parameters.

    There are some peculiarities concerning certain methods:

    ??? info "POST"

        <!-- start-POST-singularities -->
        Most POST methods implementations render their code snippets different,
        depending on *Content-Types* header, including by default some of the
        most used *Content-Types* header related behaviours:

        - The default behavior, even if you don't specify it explicitly in the
            *Content-Type* header is the generation of an
            `application/x-www-form-urlencoded` encoded request.
        - If you want to generate a ``multipart/form-data`` encoded request,
            you need to specify the files to sent using the ``files`` argument.
        - If you specifies the *Content-Type* header `application/json`, the
            parameters sent will be adjusted according to the JSON encoded POST
            request.
        - If you specifies the *Content-Type* header `text/plain`, you can only
            send one parameter and it will be adjusted accordingly following
            the implementation.
        <!-- end-POST-singularities -->

    Args:
        language (str): Programming language or plataform of the resulting code
            snippet. See [Support](/#support) to check the supported platforms
            and programming languages.
        impl (str): Implementation type used for the code snippet. It can be a
            library, a program, or a language API. See [Support](/#support)
            to check the supported implementations by language.
        method (str): HTTP method of the generated request.
        url (str, iterable, callable): URL endpoint of the generated request.
            - Defined as a string, the url will be the string itself.
            - Defined as an iterable, the url will be selected randomly
            from the iterable. Supports recursivity: until a string is not
            selected the recursion will not be stopped.
            - Defined as a callable, the url will be the returned value of
            the callable. Supports recursivity: until a string is returned the
            recursion will not be stopped.
        headers (dict): Mapping of request header names and values.
        parameters (list): List of parameters for the request. Each parameter
            must be a dictionary. This dictionary defines, for each parameter,
            what is the parameter name and how are the parameters values
            generated:

            - **name** (*str*, *list*, *function*): Parameter name. At least
                one of this or ``names`` attributes are required.
                - Defined as a string, the name will be the string itself.
                - Defined as an iterable, the name will be selected
                randomly from the iterable. Supports recursivity: until a
                string is not selected the recursion will not be stopped.
                - Defined as a callable, the name will be the returned
                value of the callable. Supports recursivity: until a string is
                returned the recursion will not be stopped.

            - **names** (*str*, *list*, *function*): Parameter name. At least
                one of this or ``name`` attributes are required.
                - Defined as an iterable, the name will be selected
                randomly from the iterable. Supports recursivity: until a
                string is not selected the recursion will not be stopped.
                - Defined as a string, must be a Python formatted module
                path following the format ``'path.to.module::function'`` and
                the return name will be used as the name for the parameter,
                which is useful if choosing a random value from a list doesn't
                fit your needs.
                - Defined as a callable, the name will be the returned
                value of the callable. Supports recursivity: until a string is
                returned the recursion will not be stopped.

            - **type** (*str*, *iterable*, *callable*): Parameter data type.
                If not defined and ``value``, ``values`` and ``faker`` are not
                defined, will be considered as a string and the value of the
                parameter will be a random word built using [faker](
                https://faker.readthedocs.io) library. For some types, other
                parameter dictionary attributes are supported, documented, if
                so, in each type. The following parameter data types are
                supported as attributes of parameters dictionaries, as well as
                their corresponding names in capital letters:

                - ``'str'``: Basic string type. Can be defined with the Python
                    builtin type ``str`` or the strings ``'str'`` and
                    ``'string'``.
                - ``'int'``: Basic integer type. Can be defined with the Python
                    builtin type ``int``, or the strings ``'int'`` and
                    ``'integer'``. As default will be an integer in the range
                    -65536 to -65536. The minimum and maximum values can be
                    defined with ``min`` and ``max`` parameter optional
                    attributes.
                - ``'float'``: Floating point number type. Can be defined
                    with the Python builtin type ``float``, or the strings
                    ``'float'`` and ``number``. As default will be in the range
                    -65536 to 65536. The minimum and maximum values
                    can be defined with ``min`` and ``max`` parameter optional
                    attributes and can be rounded with ``round`` Python builtin
                    function using ``round`` parameter optional attribute.
                - ``'bool'``: Basic boolean type. Can be defined with the
                    Python builtin type ``bool``, or the strings ``'bool'``
                    and ``'boolean'``. Returns as parameter value one of the
                    strings ``'true'`` or ``'false'``. If you pass the optional
                    parameter attribute ``null`` as ``True``, the string
                    ``'null'`` can also be returned.
                - ``'uuid'``: Unique identifier type. Can be defined with the
                    Python type ``uuid.UUID``, or the strings ``'uuid'``
                    and ``'uuid4'``. It's a unique identifier v4 encoded as
                    hexadecimal string.
                - ``'id'``: Basic integer id. It's a positive integer in the
                    range 1 to 65536. The maximum value can be defined by
                    optional ``max`` attribute.
                - ``'random'``: Random type between the available types. You
                    can define a set of possible types passing an iterable
                    to ``types`` optional parameter attribute.

                Defined as an iterable or callable, the type will be selected
                randomly from the iterable, or the returned value from the
                callable will be used. This allows you to select a random
                type from a list of custom predefined types.

            - **value** (*str*, *iterable*, *callable*): Parameter value. If
                not defined and ``type``, ``values`` and ``faker`` are not
                defined, the value of the parameter will be a random word built
                using faker library.
                - Defined as a string, the value will be the string itself.
                - Defined as an iterable, the value will be selected
                randomly from the iterable. Supports recursivity: until a
                string is not selected the recursion will not be stopped.
                - Defined as a callable, the value will be the returned
                value of the callable. Supports recursivity: until a string is
                returned the recursion will not be stopped.

            - **values** (*list*, *iterable*, *callable*): Possible parameter
                values.

                - Defined as an iterable, the value will be selected
                randomly from the iterable. Supports recursivity: until a
                string is not selected the recursion will not be stopped.
                -  Defined as a string, must be a Python formatted module
                path following the format ``'path.to.module::function'`` and
                the return value will be used as the value for the parameter,
                which is useful if choosing a random value from a list doesn't
                fit your needs.
                - Defined as a callable, the value will be the returned
                value of the callable. Supports recursivity: until a string is
                returned the recursion will not be stopped.

            - **faker** (*str*, *function*): Python formatted module path to a
                function of a Faker provider used to build the value
                randomized. Can be a standard, external provider or any
                function, but if is not a provider, ``seed`` and ``locale``
                will not have effect.
                - Defined as a string must follow the format
                ``'path.to.provider.module::function'``.

        files (dict): Mapping of files to send to URL. Only has effect for POST
            methods. If you define this argument, the `Content-Type` header of
            the request will be assumed to be `'multipart/form-data'`, but only
            will be explicitly specified in the code generated if the
            implementation needs it. Each value accepts a string, ``None`` or a
            tuple:

            + Defined as a string, must be the filepath of the file to be sent.
            + Defined as ``None``, the filepath will be randomized using
            ``faker.providers.file::file_path`` function.
            + Defined as a tuple, the first value must be the filepath of the
            file to be sent (if ``None`` will be a randomized filepath), the
            second value the content-type of the file and the third a
            dictionary of custom headers for the file.

        wrap (int): Maximum anchor of the rendered code snippet. If it exceeds
            it, the rendered code will be conveniently formatted on multiple
            lines.
        indent (str): Indentation string used in the generated code. If not
            defined, the indentation string commonly used in the implementation
            will be used.
        quote_char (str): Quotation character for strings used in the generated
            code.
        setup (bool, str): If ``True``, includes the code needed by an
            implementation to perform the request. Could be imports of
            additional modules or intialization of objects, depends on
            implementation. You can customize this snippet passing a string
            with the code snippet that you want to include.
        teardown (str): Code snippet to include after the HTTP request code.
        oneline (bool): Render the code in a single line.
        seed (int): Seed used generating random fake values of parameters.
            Useful if you want to generate the same set of values between
            multiples code snippets.
        locale (str): Locale used by [faker](https://faker.readthedocs.io)
            library to localize the faked random values for parameters.

    Raises:
        ValueError: Value is not a valid value in their context.
        TypeError: Values does not complaint with the types supported for it.
        ImportError: Python module-function path specified can not be imported
            successfully.

    Returns:
        str: HTTP request code snippet.
    '''
    _function_kwargs = {
        'parameters': parameters, 'headers': headers,
        'oneline': oneline, 'seed': seed, 'locale': locale,
        'teardown': teardown, 'wrap': wrap or float('inf')
    }
    if indent is not None:
        _function_kwargs['indent'] = indent
    if quote_char is not None:
        _function_kwargs['quote_char'] = quote_char
    if method.lower() == 'post':
        _function_kwargs['files'] = files
    if setup is not None:
        _function_kwargs['setup'] = setup
    kwargs.update(_function_kwargs)
    return get_func_by_lang_impl_method(
        language=language, impl=impl, method=method
    )(lazy_string(url), **kwargs)


def generate_http_request_md_fenced_code_block(language=None,
                                               fence_string='```',
                                               **kwargs):
    """Wraps [``generate_http_request_code``](#generate_http_request_code)
    function result in a Markdown fenced code block.

    Args:
        fence_string (str): Code block fence string used wrapping the code.
            It does not perform any check about if the fenced string is a
            "valid" markdown code block fence string.
        **kwargs: All other optional arguments are passed to
            [``generate_http_request_code``](#generate_http_request_code)
            function.

    Examples:
        >>> generate_http_request_md_fenced_code_block(setup=False)
        "```python\\nreq = requests.get('http://localhost')\\n```"

        >>> generate_http_request_md_fenced_code_block(fence_string='~~~',
        ...                                            setup=False)
        "~~~python\\nreq = requests.get('http://localhost')\\n~~~"

    Returns:
        str: Fenced code block with HTTP request code snippet inside.
    """
    return '%(fence_string)s%(language)s\n%(render)s\n%(fence_string)s' % {
        'language': language if language else DEFAULT_LANGUAGE,
        'render': generate_http_request_code(language=language, **kwargs),
        'fence_string': fence_string
    }



'''Common exceptions for http-request-codegen.'''


def raise_post_text_plain_n_parameters_not_1(n_parameters,
                                             exc_type=ValueError):
    '''Raises a ``ValueError`` with a message indicating that the user must
    pass only one parameter attempting to generate ``text/plain`` content-type
    encoded POST request.

    Args:
        n_parameters (int): Number of parameters provided by the user.
        exc_type (type): Exception type to raise.

    Examples:
        >>> raise_post_text_plain_n_parameters_not_1(5)
        Traceback (most recent call last):
        ...
        ValueError: You can only send one parameter making a POST ..., got 5
    '''
    raise exc_type(
        ('You can only send one parameter making a POST request encoded'
         ' as \'text/plain\', got %d') % n_parameters)



'''Utilities for Bash HTTP request generators.'''

from http_request_codegen.generators.python._utils import (  # noqa: F401
    escape_by_quote
)


DEFAULT_INDENT = '    '
DEFAULT_QUOTE_CHAR = '"'
DEFAULT_WRAP = 80



'''Parameter value formatter factory.'''

import importlib
import random
import uuid
from functools import lru_cache

from faker import Faker
from faker.providers import lorem as faker_lorem_provider

from http_request_codegen.hrc_meta import CallableTypes
from http_request_codegen.hrc_string import lazy_string


@lru_cache(maxsize=32)
def _instanciate_faker(seed=None, locale=None):
    if seed is not None:
        Faker.seed(seed)
    faker = Faker(locale)
    return faker


def lazy_name_by_parameter(parameter_data, seed=None):
    '''Given a dictionary of parameter options, returns the corresponding
    parameter name built following the rules listed in ``parameters`` argument
    of [``generate_http_request_code``](#generate_http_request_code) function
    documentation.

    The strategy of name building is to check next attributes in given order:

    - ``name``
    - ``names``

    You can use this function to build the parameters at lower level. This can
    be used, for example, to append the parameters to an URL generating GET
    method code snippets if an implementation does by building the parameters
    as arguments of a function.

    Args:
        parameter_data (dict): Parameter specification data. It's defined at
            **name** and **names** sections of ``parameters`` argument of
            [``generate_http_request_code``](#generate_http_request_code)
            function documentation.
        seed (int): Seed using randomizing names.

    Raises:
        ValueError: if none of the name or names attributes are defined inside
            ``parameter_data`` dictionary.

    Examples:
        >>> lazy_name_by_parameter({'name': 'foo'})
        'foo'

        >>> result = lazy_name_by_parameter({'names': ['foo', 'bar', 'baz']})
        >>> result in ['foo', 'bar', 'baz']
        True

    Returns:
        str: Parameter name.
    '''
    if 'name' in parameter_data:
        return lazy_string(parameter_data['name'], seed=seed)
    elif 'names' in parameter_data:
        try:
            return lazy_string(parameter_data['names'],
                               seed=seed,
                               string_func_path=True)
        except (ModuleNotFoundError, ImportError):
            raise ValueError(
                ('\'names\' \'%s\' attribute of parameter is pointing to an'
                 ' inexistent Python callable object') % (
                     parameter_data['names']))

        return lazy_string(parameter_data['names'],
                           seed=seed, string_func_path=True)
    raise ValueError(('Parameter must contain \'name\' or \'names\''
                      ' attribute, got "%s"') % str(parameter_data))


def lazy_value_by_parameter(parameter_data, seed=None, locale=None):
    '''Given a dictionary of parameter options, returns the corresponding value
    built following the rules listed in ``parameters`` argument of
    [``generate_http_request_code``](#generate_http_request_code) function
    documentation.

    For example, giving ``{\'type\': int}`` as input, the output will be a
    random number as string.

    The strategy of value building is to check next attributes in given order:

    - ``'value'``
    - ``'values'``
    - ``'faker'``
    - ``'type'``

    If any of the previous attributes are passed will be treated as if
    ``{\'type\': str}`` has been passed, returning a random word.

    You can use this function to build the parameters at lower level. This can
    be used, for example, to append the parameters to an URL generating GET
    method code snippets if an implementation does by building the parameters
    as arguments of a function.

    Args:
        parameter_data (dict): Parameter specification data. It's defined at
            **type**, **value**, **values** and **faker** sections of
            ``parameters`` argument as is defined at
            [``generate_http_request_code``](#generate_http_request_code)
            function documentation.
        seed (int): Seed using randomizing values.
        locale (str): Locale used for ``faker`` providers.

    Examples:
        >>> lazy_value_by_parameter({'value': 'foo'})
        'foo'

        >>> result = lazy_value_by_parameter({'values': ['foo', 'bar', 'baz']})
        >>> result in ['foo', 'bar', 'baz']
        True

        >>> result = lazy_value_by_parameter({'type': 'int'})
        >>> result.replace('.', '', 1).lstrip('-').isnumeric() and \\
        ...     isinstance(result, str)
        True

    Raises:
        ImportError: if ``'values'`` attribute value points to an inexistent
            Python object.
        TypeError: if ``'faker'`` attribute value does not contains an string
            or Python callable object, or if the ``'type'`` attribute value
            does not support the defined type.
        ImportError: if ``'faker'`` attribute value, when passed as string,
            points to an inexistent Python object.

    Returns:
        str: Parameter value.
    '''
    if 'value' in parameter_data:
        return lazy_string(parameter_data['value'], seed=seed)
    elif 'values' in parameter_data:
        try:
            return lazy_string(parameter_data['values'],
                               seed=seed,
                               string_func_path=True)
        except (ModuleNotFoundError, ImportError):
            raise ImportError(
                ('\'values\' \'%s\' attribute of parameter \'%s\' is'
                 ' pointing to an inexistent Python callable object') % (
                     parameter_data['values'], parameter_data['name']))
    elif 'faker' in parameter_data:
        if isinstance(parameter_data['faker'], str):
            # Search provider by string
            provider_mod_name, func_name = parameter_data['faker'].split("::")
            mod = importlib.import_module(provider_mod_name)
            faker = _instanciate_faker(seed=seed, locale=locale)
            faker.add_provider(mod)
            return getattr(faker, func_name)()
        elif isinstance(parameter_data['faker'], CallableTypes):
            faker = _instanciate_faker(seed=seed, locale=locale)
            faker.add_provider(parameter_data['faker'].__module__)
            return getattr(faker, parameter_data['faker'].__name__)()
        raise TypeError(
            ('\'faker\' \'%s\' attribute of parameter \'%s\' must be an'
             ' instance of \'str\' or \'callable\'') % (
                 str(parameter_data['faker']), parameter_data['name']))
    if 'type' not in parameter_data:
        _type = 'str'
    else:
        _type = lazy_string(parameter_data['type'], seed=seed)
        if isinstance(_type, str):
            _type = _type.lower()
    if _type in ('str', 'string', str):
        faker = _instanciate_faker(seed=seed, locale=locale)
        faker.add_provider(faker_lorem_provider)
        return faker.word()
    elif _type in ('int', 'integer', int):
        if seed is not None:
            random.seed(seed)
        # Document max and min in public API
        _max = 65536 if 'max' not in parameter_data \
            else parameter_data['max']
        _min = -65536 if 'min' not in parameter_data \
            else parameter_data['min']
        return str(random.randint(_min, _max))
    elif _type in ('float', 'number', float):
        if seed is not None:
            random.seed(seed)
        # Document step in public API
        _max = 65536 if 'max' not in parameter_data \
            else parameter_data['max']
        _min = -65536 if 'min' not in parameter_data \
            else parameter_data['min']
        response = random.uniform(_min, _max)
        if 'round' in parameter_data:
            response = round(response, parameter_data['round'])
        return str(response)
    elif _type in ('bool', 'boolean', bool):
        _possibles = ['true', 'false']
        if parameter_data.get('null'):
            _possibles.append('null')
        if seed is not None:
            random.seed(seed)
        return random.choice(_possibles)
    elif _type in ('uuid', 'uuid4', uuid.UUID):
        return _instanciate_faker(
            seed=seed, locale=locale).uuid4(cast_to=None).hex
    elif _type in ('id', 'identifier'):
        if seed is not None:
            random.seed(seed)
        _max = 65536 if 'max' not in parameter_data \
            else parameter_data['max']
        return str(random.randint(1, _max))
    elif _type == 'random':
        if seed is not None:
            random.seed(seed)
        if 'types' in parameter_data:
            _possible = lazy_string(parameter_data['types'], seed=seed)
        else:
            _possible = [
                'str', 'int', 'float', 'bool', 'uuid', 'id', 'file']
        parameter_data['type'] = random.choice(_possible)
        return lazy_value_by_parameter(parameter_data['type'], seed=seed)
    raise TypeError(
        ('Data type \'%s\' of parameter \'%s\' not supported') % (
            parameter_data['type'], parameter_data['name']))




import inspect
from types import FunctionType, LambdaType, MethodType


CallableTypes = (FunctionType, MethodType, LambdaType)


def function_has_kwarg(func, kwarg_name):
    '''Discovers if a given function has the argument ``kwarg_name`` in their
    definition. The check is not exhaustive because if an argument of the
    function contains the value ``'{kwarg_name}='`` will be falsely discovered.
    Useful only if you are sure that the function definition does not contains
    such argument value.

    It works also for module, class, method, function, traceback, frame, or
    code objects, but it only makes sense to use it for methods or functions.

    Args:
        func (function): Function to check if their keyword arguments contains
            the given ``kwarg_name``. If is not a module, class, method,
            function, traceback, frame, or code object, a TypeError argument
            will be raised.
        kwarg_name (str): Keyword argument name to check if it's defined as a
            keyword argument of the function ``func``. Does not trigger errors
            if the value passed to this argument can be casted to strings using
            string templates with ``'%s'`` syntax.

    Raises:
        TypeError: if  ``func`` is not a Python function.

    Examples:
        >>> def function_without_foo_kwarg(foo):
        ...     pass
        >>> function_has_kwarg(function_without_foo_kwarg, 'foo')
        False

        >>> def function_with_foo_kwarg(foo=None):
        ...     pass
        >>> function_has_kwarg(function_with_foo_kwarg, 'foo')
        True

        >>> def function_with_arg_valued_like_foo_kwarg_def(bar='foo=1'):
        ...     pass
        >>> function_has_kwarg(function_with_arg_valued_like_foo_kwarg_def,
        ...                    'foo')
        True

        >>> function_has_kwarg('bar', 'foo')
        Traceback (most recent call last):
          ...
        TypeError: function or method was expected, got str

        >>> def function_with_kwarg_def_123_valued(foo=123):
        ...     pass
        >>> function_has_kwarg(function_with_kwarg_def_123_valued, 123)
        False

    Returns:
        bool: ``True`` if the function contains such keyword argument name or
            ``False`` otherwise.
    '''
    response = False
    _inside_func_def = False
    try:
        func_source = inspect.getsource(func)
    except TypeError:
        raise TypeError(
            ('function or method was expected, got %s') % type(func).__name__)

    for line in func_source.split("\n"):
        if not _inside_func_def and line.lstrip().startswith("def "):
            _inside_func_def = True

        # Would be rare if a kwarg as a string with default value containing
        # 'kwarg_name='. For this library we don't need this check
        if _inside_func_def and (('%s=' % kwarg_name) in line):
            response = True
            break

        if _inside_func_def and line.rstrip().endswith("):"):
            break
    return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import re
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup


URL = "https://github.com/mondeja/http-request-codegen"
EMAIL = "mondejar1994@gmail.com"
AUTHOR = "Álvaro Mondéjar Rubio"

REQUIRES_PYTHON = ">=3"
REQUIRED = [
    'faker>=4.0.0',  # Python3 only support
]
LINT_EXTRAS = [
    'flake8==3.8.4',
    'flake8-print==4.0.0',
    'flake8-implicit-str-concat==0.2.0',
    'isort==5.7.0',
    'yamllint==1.25.0',
]
TEST_EXTRAS = [
    'pytest==6.2.1',
    'pytest-cov==2.10.1',
    'inflection==0.5.1',
    'flask==1.1.2',
    'requests==2.25.1',
]
DOC_EXTRAS = [
    'mkdocs==1.1.2',
    'mkdocs-material==6.2.3',
    'mkdocs_macros_plugin==0.5.0',
    'mkdocs-minify-plugin==0.3.0',
    'mkdocs-include-markdown-plugin==2.7.1',
    'mkdocstrings==0.14.0',
    'mkdocs-exclude==1.0.2',
]
DEV_EXTRAS = [
    'twine==3.3.0',
    'bump2version==1.0.1',
    'pre-commit==2.9.3',
] + DOC_EXTRAS + TEST_EXTRAS + LINT_EXTRAS

HERE = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = "\n" + f.read()

ABOUT = {}
INIT_FILEPATH = os.path.join(HERE, "http_request_codegen", "__init__.py")
with io.open(INIT_FILEPATH, encoding="utf-8") as f:
    content = f.read()
    ABOUT["__title__"] = \
        re.search(r"__title__\s=\s[\"']([^\"']+)[\"']", content).group(1)
    ABOUT["__version__"] = \
        re.search(r"__version__\s=\s[\"']([^\"']+)[\"']", content).group(1)
    ABOUT["__description__"] = \
        re.search(r"__description__\s=\s[\"']([^\"']+)[\"']", content).group(1)


class UploadCommand(Command):
    '''Support setup.py upload.'''

    description = "Build and publish the package."
    user_options = [
        ("test", None, "Specify if you want to test your upload to Pypi."),
    ]

    @staticmethod
    def status(s):
        '''Prints things in bold.'''
        sys.stdout.write("\033[1m{0}\033[0m\n".format(s))

    def initialize_options(self):
        self.test = None

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(HERE, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(
            sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        cmd = "twine upload%s dist/*" % (
            " --repository-url https://test.pypi.org/legacy/" if self.test
            else ""
        )
        os.system(cmd)
        sys.exit()


setup(
    name=ABOUT["__title__"],
    version=ABOUT["__version__"],
    description=ABOUT["__description__"],
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests"]),
    install_requires=REQUIRED,
    extras_require={
        "dev": DEV_EXTRAS,
        "test": TEST_EXTRAS,
        "lint": LINT_EXTRAS,
        "doc": DOC_EXTRAS,
    },
    include_package_data=True,
    license="BSD License",
    classifiers=[
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Documentation",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    cmdclass={
        "upload": UploadCommand,
    },
)



'''Language-implementation-method factory.'''

import importlib
import os
from functools import lru_cache

from http_request_codegen.hrc_http import HTTP_METHODS


DEFAULT_LANGUAGE = 'python'
DEFAULT_IMPLEMENTATION = 'requests'

GENERATORS_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'generators'))


@lru_cache(maxsize=1)
def get_generators_modules_by_lang_impl():
    response = {}

    for fname in os.listdir(GENERATORS_DIR):
        if fname.startswith('_'):
            continue
        fpath = os.path.join(GENERATORS_DIR, fname)
        if not os.path.isdir(fpath):
            continue
        response[fname] = {}

        for modfname in os.listdir(fpath):
            if modfname.startswith('_'):
                continue
            modname = modfname.rstrip('.py')
            response[fname][modname] = (
                'http_request_codegen.generators.%s.%s'
            ) % (fname, modname)
    return response


@lru_cache(maxsize=32)
def get_func_by_lang_impl_method(language=None, impl=None, method=None):
    generators_by_lang_impl = get_generators_modules_by_lang_impl()

    if language is None:
        if impl is None:
            _language = 'python'
        else:
            _language = None
            for __lang, __imps in generators_by_lang_impl.items():
                for __imp in __imps.keys():
                    if __imp == impl:
                        _language = __lang
                        break
                if _language is not None:
                    break
            if _language is None:
                _language = 'python'
    else:
        _language = language
    try:
        lang_impls = generators_by_lang_impl[_language]
    except KeyError:
        raise ValueError(('The language \'%s\' is not implemented in'
                          ' http-request-codegen.') % language)
    if impl is None:
        _impl = list(lang_impls.keys())[0]
    else:
        _impl = impl

    try:
        impl_modpath = lang_impls[_impl]
    except KeyError:
        raise ValueError(
            ('The implementation \'%s\' is not implemented'
             ' for the language \'%s\' in http-request-codegen.') % (
                 impl, _language))

    module = importlib.import_module(impl_modpath)

    if method is None:
        _method = 'get'
    else:
        _method = method.lower()
    try:
        func = getattr(module, _method)
    except AttributeError:
        if method.upper() not in HTTP_METHODS:
            raise ValueError('Invalid HTTP method \'%s\'' % method.upper())
        raise ValueError(
            ('The implementation \'%s\' of the language \'%s\' does not'
             ' support HTTP %s methods.') % (impl, _language, _method.upper())
        )
    return func



'''Utilities for Python HTTP request generators.'''

from http_request_codegen.hrc_string import (
    escape_double_quote,
    escape_single_quote,
    lazy_escape_quote_func_by_quote_char
)


DEFAULT_INDENT = '    '
DEFAULT_QUOTE_CHAR = "'"
DEFAULT_WRAP = 80

VALID_QUOTE_CHARS = '"\''


def validate_python_identifier(value):
    '''Validates if a string is a valid Python identifier. If not, raises
    a ValueError.

    Args:
        value (str): Value to validate.

    Raises:
        ValueError: if the string is not a valid Python identifier.

    Examples:
        >>> validate_python_identifier('_hello')
        >>> validate_python_identifier('123')
        Traceback (most recent call last):
          ...
        ValueError: '123' is not a valid Python identifier

        >>> validate_python_identifier(123)
        Traceback (most recent call last):
          ...
        ValueError: '123' is not a valid Python identifier
    '''
    try:
        if not value.isidentifier():
            raise ValueError('\'%s\' is not a valid Python identifier' % value)
    except AttributeError:
        raise ValueError(
            '\'%s\' is not a valid Python identifier' % str(value))


def validate_quote_character(char):
    '''Validates if a character is a valid Python string quotation character.

    Args:
        char (str): Character to validate.

    Raises:
        ValueError: if the character is not a valid Python string quotation
            character.

    Examples:
        >>> validate_quote_character('"')
        >>> validate_quote_character('1')
        Traceback (most recent call last):
          ...
        ValueError: '1' is not a valid Python quotation character
    '''
    if char not in VALID_QUOTE_CHARS:
        raise ValueError('\'%s\' is not a valid Python quotation character' % (
                         char))


def escape_quote_func_by_quote_char(char):
    '''Get a function that can escape a string quotation character given
    the character. This works as a factory for quotation string characters
    escapes functions.

    Args:
        char (str): Character that the returned function is able to escape.

    Raises:
        ValueError: if the character passed is not a valid Python string
            quotation character.

    Examples:
        >>> func = escape_quote_func_by_quote_char("'")
        >>> print(func("String that must be ' escaped."))
        String that must be \\' escaped.

        >>> escape_quote_func_by_quote_char("?")
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Python quotation character

    Returns:
        function: Function that can escape the character passed as argument.
    '''
    return lazy_escape_quote_func_by_quote_char(
        char,
        replacers_funcs={
            '"': escape_double_quote,
            '\'': escape_single_quote
        },
        error_msg_schema=('%(quote_char)s is an invalid Python quotation'
                          ' character')
    )


def escape_by_quote(string, char):
    '''Escapes the quote characters of a string.

    Args:
        string (str): String whose qutation characters must be escaped.
        char (str): Quotation character to escape.

    Raises:
        ValueError: If the character to escape is not a valid Python string
            quotation character.
        TypeError: If the value to escape is not a string.

    Examples:
        >>> print(escape_by_quote("I need single quote ' escape.", "'"))
        I need single quote \\' escape.

        >>> print(escape_by_quote('I need double quote " escape.', '"'))
        I need double quote \\" escape.

        >>> escape_by_quote(12, '"')
        Traceback (most recent call last):
          ...
        TypeError: The value '12' can not be escaped because is not a string

        >>> escape_by_quote('I need escape of ? character.', '?')
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Python quotation character

        >>> escape_by_quote(12, '?')
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Python quotation character

    Returns:
        str: The original string with the specified character escaped.
    '''
    return escape_quote_func_by_quote_char(char)(string)


def kwarg_definition(kwarg_name, value, indent=DEFAULT_INDENT, **kwargs):
    '''Returns the code of any value passed as Python keyword argument to
    a function. The indentation used before the keyword argument is 4 spaces
    by default.

    Args:
        kwarg_name (str): Name of the argument. Must be a valid Python
            identifier or a `ValueError` will be raised.
        value (object): Value of the argument to be reproduced.
        **kwargs: Arguments passed to the functions
            ``kwarg_definition_dict_valued`` and
            ``kwarg_definition_str_valued``.

    Raises:
        ValueError: If the argument ``kwarg_name`` is not a valid Python
            identifier.

    Examples:
        >>> print(kwarg_definition('foo', 'bar'))
        foo='bar'
        >>> print(kwarg_definition('foo', 1))
        foo=1
        >>> print(kwarg_definition('foo', dict(bar='baz')))
            foo={
                'bar': 'baz'
            }
        >>> print(kwarg_definition('foo', dict(bar=1, baz=2)))
            foo={
                'bar': 1,
                'baz': 2
            }

        >>> print(kwarg_definition(123, 'a'))
        Traceback (most recent call last):
          ...
        ValueError: '123' is not a valid Python identifier

    Returns:
        str: Code needed to pass an optional keyword argument to a function
            in Python
    '''
    # Invalid Python identifiers can't be defined as keyword arguments
    validate_python_identifier(kwarg_name)

    if isinstance(value, dict):
        return kwarg_definition_dict_valued(kwarg_name, value, indent=indent,
                                            _validate_identifier=False,
                                            **kwargs)
    elif isinstance(value, str):
        return kwarg_definition_str_valued(kwarg_name, value, indent=indent,
                                           _validate_identifier=False,
                                           **kwargs)
    return '%(kwarg_name)s=%(value)s' % {
        'kwarg_name': kwarg_name,
        'value': repr(value),
    }


def kwarg_definition_str_valued(kwarg_name, string,
                                quote_char=DEFAULT_QUOTE_CHAR,
                                indent=DEFAULT_INDENT,
                                wrap=DEFAULT_WRAP,
                                _validate_identifier=True):
    '''Returns the code of a string passed as Python keyword argument to
    a function.

    Args:
        kwarg_name (str): Name of the argument. Must be a valid Python
            identifier or a ``ValueError`` will be raised.
        dictionary (str): Dictionary of keys-values shown as passed to the
            argument of the function.
        quote_char (str): Quotation mark character used in strings. Must be
            a valid Python quotation character or a `ValueError` will be
            raised.
        indent (str): Indentation string.
        _validate_identifier (bool): Indicates if the argument ``kwarg_name``
            must be evaluated as a valid Python identifier.

    Raises:
        ValueError: if the argument ``kwarg_name`` is not a valid Python
            identifier and ``_validate_identifier`` is ``True`` or if the
            ``quote_char`` argument is not a valid Python string quotation
            character.
        TypeError: if the argument ``quote_char`` is not a string.

    Examples:
        >>> print(kwarg_definition_str_valued('foo', 'bar'))
        foo='bar'
        >>> print(kwarg_definition_str_valued('foo', 'bar', quote_char='"',
        ...                                   indent=''))
        foo="bar"
        >>> print(kwarg_definition_str_valued('123', 'bar'))
        Traceback (most recent call last):
          ...
        ValueError: '123' is not a valid Python identifier
        >>> print(kwarg_definition_str_valued('foo', 'bar', quote_char='?'))
        Traceback (most recent call last):
          ...
        ValueError: '?' is an invalid Python quotation character
        >>> print(kwarg_definition_str_valued('foo', 'bar', quote_char=123))
        Traceback (most recent call last):
          ...
        ValueError: '123' is an invalid Python quotation character

    Returns:
        str: Reproduction of the kwarg defined as a Python keyword argument
            passed to a function.
    '''
    if _validate_identifier:
        validate_python_identifier(kwarg_name)

    return '%(kwarg_name)s=%(value)s' % {
        'kwarg_name': kwarg_name,
        'value': str_definition(
            string,
            indent=' ' * (len(kwarg_name) + len(indent) + 1),
            quote_char=quote_char,
            wrap=wrap),
    }


def kwarg_definition_dict_valued(kwarg_name, dictionary,
                                 quote_char=DEFAULT_QUOTE_CHAR,
                                 indent=DEFAULT_INDENT, indent_depth=1,
                                 newline='\n', wrap=DEFAULT_WRAP,
                                 _validate_identifier=True, _escape_keys=True,
                                 _escape_values=True):
    '''Returns the code of a dictionary passed as Python keyword argument
    to a function.

    Args:
        kwarg_name (str): Name of the argument. Must be a valid Python
            identifier or a `ValueError` will be raised.
        dictionary (str): Dictionary of keys-values shown as passed to the
            argument of the function.
        quote_char (str): Quotation mark character used in strings. Must be
            a valid Python string quotation character or a `ValueError` will be
            raised.
        indent (str): Indentation string.
        indent_depth (int): First indentation level.
        newline (str): Newline string.
        _validate_identifier (bool): Indicates if the argument ``kwarg_name``
            must be evaluated as a valid Python identifier.
        _escape_keys (bool): Indicates if the keys of the reproducted
            dictionary must be escaped by the given ``quote_char`` argument.
        _escape_values (bool): Indicates if the values of the reproducted
            dictionary must be escaped by the given ``quote_char`` argument.

    Examples:
        >>> print(
        ...     kwarg_definition_dict_valued(
        ...         'headers',
        ...         {'Authorization': '__token__',
        ...          'Content-Type': 'application/json'}
        ...     )
        ... )
            headers={
                'Authorization': '__token__',
                'Content-Type': 'application/json'
            }

        >>> print(kwarg_definition_dict_valued('foo', dict(bar='baz'),
        ...                                    quote_char='"'))
            foo={
                "bar": "baz"
            }

        >>> print(kwarg_definition_dict_valued('foo', {'bar': 1},
        ...                                    quote_char="'", indent_depth=0))
        foo={
            'bar': 1
        }

        >>> print(kwarg_definition_dict_valued('foo', {'bar': 'baz'},
        ...                                    newline='', indent='',
        ...                                    indent_depth=0))
        foo={'bar': 'baz'}

    Returns:
        str: Python keyword argument with dictionary as value been called
            by a Python function.
    '''
    if _validate_identifier:
        validate_python_identifier(kwarg_name)

    return '%(indent)s%(kwarg_name)s=%(dictionary)s' % {
        'indent': indent * indent_depth,
        'kwarg_name': kwarg_name,
        'dictionary': dict_definition(
            dictionary,
            indent=indent,
            indent_depth=indent_depth,
            quote_char=quote_char,
            wrap=wrap,
            newline=newline,
            _escape_keys=_escape_keys,
            _escape_values=_escape_values
        )
    }


def str_definition(string, indent=DEFAULT_INDENT,
                   quote_char=DEFAULT_QUOTE_CHAR, wrap=DEFAULT_WRAP,
                   _escape=True):
    '''Creates a definition of a Python string, multilining it if neccesary.
    Does not handle spaces at all wrapping it, so it's useful to efficiently
    wrap non-spaced strings like URLs.

    Args:
        string (str): String to be reproducted.
        indent (str): Indentation string. This defines the space at the left of
            the string in the code with respect to column 0 of the code.
        quote_char (str): Quotation mark character used in the string. Must be
            a valid Python string quotation character or a `ValueError` will be
            raised.
        wrap (int): Maximum anchor in which the string will be wrapped in
            multiples lines with respect to column 0 of the code.
        _escape (bool): Defines if the reproducted string must be escaped by
            the given ``quote_char`` argument.

    Raises:
        ValueError: if the argument ``quote_char`` is not a valid Python string
            quotation character.

    Examples:

        >>> print(str_definition(
        ...           'Lorem Ipsum es simplemente el texto de'
        ...           + ' relleno de las imprentas y archivos de texto.',
        ...           wrap=70, indent=''))
        ('Lorem Ipsum es simplemente el texto de relleno de las imprentas y '
         'archivos de texto.')

        >>> print(str_definition('123', wrap=1, quote_char='"', indent=''))
        ("1"
         "2"
         "3"
        )
        >>> print(str_definition('123', wrap=5, quote_char='"', indent=''))
        ("1"
         "2"
         "3")
        >>> print(str_definition('123', wrap=6, quote_char='"', indent=''))
        "123"

    Returns:
        str: String reproducted in multiples lines wrapped by ``(`` and ``)``
            characters in multiples lines or simply defined in a single line.
    '''
    string_escaped = string if not _escape else \
        escape_by_quote(string, quote_char)

    if len(string) + len(indent) + len(quote_char) * 2 < wrap:
        return '%(quote_char)s%(value)s%(quote_char)s' % {
            'quote_char': quote_char,
            'value': string_escaped,
        }

    indent_length = len(indent)
    response = '(%(quote_char)s' % {
        'quote_char': quote_char
    }
    _chars_in_current_line = len(response) + indent_length
    for i, ch in enumerate(string):
        response += ch
        _chars_in_current_line += 1
        if _chars_in_current_line >= wrap - 2:
            if i >= len(string) - 1:
                break
            response += '%(quote_char)s\n%(indent)s %(quote_char)s' % {
                'quote_char': quote_char,
                'indent': indent,
            }
            _chars_in_current_line = 2 + indent_length  # ' ('
    response += '%(quote_char)s%(newline)s%(indent)s)' % {
        'newline': '\n' if _chars_in_current_line >= wrap - 1 else '',
        'indent': indent if _chars_in_current_line >= wrap - 1 else '',
        'quote_char': quote_char,
    }
    return response


def dict_definition(dictionary, indent=DEFAULT_INDENT, indent_depth=0,
                    quote_char=DEFAULT_QUOTE_CHAR, wrap=DEFAULT_WRAP,
                    newline='\n', _escape_keys=True, _escape_values=True,
                    _str_definition_func=str_definition):
    '''Creates a definition of a Python dictionary.

    Args:
        dictionary (str): Dictionary that will be defined as Python code.
        indent (str): Indentation used for the keys and values.
        indent_depth (int): Number of levels of indentation.
        quote_char (str): Python string quotation character used.
        wrap (int): Maximum anchor of the code. If it exceeds it, it will be
            wrapped in multiple lines.
        newline (str): Newline character.
        _escape_keys (bool): If ``True`` the keys of the dictionary will be
            escaped against the string of ``quote_char`` argument.
        _escape_values (bool): If ``True`` the values of the dictionary will be
            escaped against the string of ``quote_char`` argument.

    Examples:
        >>> print(dict_definition({'foo': 'bar'}))
        {
            'foo': 'bar'
        }

    Returns:
        str: Definition of the dictionary.
    '''
    if _escape_keys:
        escape_quote_func = escape_quote_func_by_quote_char(quote_char)

    response = '{%(newline)s' % {'newline': newline}
    for i, (key, value) in enumerate(dictionary.items()):
        _key = key if not _escape_keys else escape_quote_func(key)
        if isinstance(value, str):
            _indent = indent * indent_depth * 2 + ' ' * (len(_key) + 4)
            _value = _str_definition_func(value, indent=_indent,
                                          quote_char=quote_char, wrap=wrap,
                                          _escape=_escape_values)
        else:
            _value = str(value)
        response += ('%(indents)s%(quote_char)s%(key)s%(quote_char)s:'
                     ' %(value)s%(comma)s%(newline)s') % {
            'indents': indent * (indent_depth + 1),
            'key': _key,
            'value': _value,
            'quote_char': quote_char,
            'newline': newline,
            'comma': ',' if i < len(dictionary) - 1 else '',
        }
    response += '%(indent)s}' % {
        'indent': (indent * indent_depth) if dictionary else ''
    }
    return response
